// recognizer.h

#ifdef _MSC_VER
#define NOMINMAX
#endif

#include "recognizer3.h"

namespace kaldi {
namespace nnet3 {
NnetRecognizer::NnetRecognizer(
        kaldi::nnet3::AmNnetSimple& nnet,
        kaldi::nnet3::AmNnetSimple& cmb_nnet,
        const kaldi::TransitionModel& trans_model,
        const kaldi::TransitionModel& cmb_trans_model,
        kaldi::nnet3::NnetSimpleLoopedComputationOptions opt,
        kaldi::nnet3::DecodableNnetSimpleLoopedInfo &decodable_info,
        kaldi::nnet3::DecodableNnetSimpleLoopedInfo &cmb_decodable_info,
        kaldi::OnlineNnet2FeaturePipelineInfo *feature_info,
        std::shared_ptr<kaldi::OnlineFeatureMfcnet>& features,
        std::shared_ptr<kaldi::OnlineFeatureInterface>& features_if,
        const fst::Fst<fst::StdArc>* decode_fst,
        const fst::Fst<fst::StdArc>* cmb_decode_fst,
        const kaldi::LatticeFasterDecoderConfig& decoder_config,
        const fst::SymbolTable* phone_syms,
        const fst::SymbolTable* cmb_phone_syms,
        const fst::SymbolTable* word_syms,
        const fst::SymbolTable* cmb_word_syms,
        const kaldi::WordAlignLatticeLexiconInfo& lexicon_info,
        const kaldi::WordAlignLatticeLexiconInfo& cmb_lexicon_info,
        std::string& lm_name,
        int emulate_julius,
        int nbest_count,
        int nbest_for_lat_combine,
        int progress_output_interval,
        std::vector<float>& combine_lat_weights,
        bool enable_print_result,
        bool enable_print_progress,
        bool enable_print_sentence,
        bool enable_print_lat_combine,
        bool enable_lat_combine,
        std::shared_ptr<result_server> result_client,
        std::shared_ptr<resource_monitor> rm):
    id_(features->GetID()),
    trans_model_(trans_model), cmb_trans_model_(cmb_trans_model), opt_(opt),
    decodable_info_(decodable_info), cmb_decodable_info_(cmb_decodable_info),
    feature_info_(feature_info), features_(features), features_if_(features_if),
    decoder_(*decode_fst, decoder_config), cmb_decoder_(*cmb_decode_fst, decoder_config),
    phone_syms_(phone_syms), cmb_phone_syms_(cmb_phone_syms),
    word_syms_(word_syms), cmb_word_syms_(cmb_word_syms),
    lexicon_info_(lexicon_info), cmb_lexicon_info_(cmb_lexicon_info),
    result_client_(result_client),
    rm_(rm),
    lm_name_(lm_name),
    emulate_julius_(emulate_julius),
    nbest_count_(nbest_count), nbest_for_lat_combine_(nbest_for_lat_combine),
    progress_output_interval_(progress_output_interval),
    combine_lat_weights_(combine_lat_weights),
    enable_print_result_(enable_print_result),
    enable_print_progress_(enable_print_progress),
    enable_print_sentence_(enable_print_sentence),
    enable_print_lat_combine_(enable_print_lat_combine),
    enable_lat_combine_(enable_lat_combine),
    features_input_is_finished(false){
  //if(features_){ std::string id_ = features_->GetID(); }
  // remove path prefix
#ifdef _MSC_VER
  std::string::size_type pos = lm_name.rfind("\\");
#else
  std::string::size_type pos = lm_name.rfind("/");
#endif
  lm_name_ = (pos != std::string::npos) ? lm_name.substr(++pos) : lm_name;
  emulate_julius_ = (emulate_julius != 0) ? 1 : 0;
  delimiter_sym_ = (emulate_julius_) ? ".\n" : "";

  adaptation_state = new kaldi::OnlineIvectorExtractorAdaptationState(feature_info->ivector_extractor_info);
  silence_weighting = new kaldi::OnlineSilenceWeighting(trans_model, feature_info->silence_weighting_config, opt.frame_subsampling_factor);
  feature_pipeline = new kaldi::OnlineNnet2FeaturePipelineMod(*feature_info, features_if_);
  feature_pipeline->SetAdaptationState(*adaptation_state);

  decodable_ = new kaldi::nnet3::DecodableNnet3SimpleOnlineForHark(decodable_info_, trans_model_, opt_, feature_pipeline->InputFeature(), feature_pipeline->IvectorFeature());
  cmb_decodable_ = new kaldi::nnet3::DecodableNnet3SimpleOnlineForHark(cmb_decodable_info_, cmb_trans_model_, opt_, feature_pipeline->InputFeature(), feature_pipeline->IvectorFeature());
}

std::shared_ptr<OnlineFeatureInterface> NnetRecognizer::GetOnlineFeatureInterface()
{
  return features_if_;
}

void NnetRecognizer::Recognize()
{
  if(features_){
    features_->Initialize();
    std::ostringstream ss_sock;
    if(features_->is_ex_mfcnet_input()){
      ss_sock << std::fixed << std::setprecision(6)
              << "<SOURCEINFO CLIENTID=\"" << features_->GetUtteranceInfo().client_id
              <<          "\" MICINDEX=\"" << features_->GetUtteranceInfo().mic_index
              <<          "\" SOURCEID=\"" << features_->GetUtteranceInfo().source_id
              <<             "\" SUBID=\"" << features_->GetUtteranceInfo().sub_id
              <<           "\" AZIMUTH=\"" << features_->GetUtteranceInfo().azimuth
              <<         "\" ELEVATION=\"" << features_->GetUtteranceInfo().elevation
              <<               "\" SEC=\"" << features_->GetUtteranceInfo().tv_sec
              <<              "\" USEC=\"" << features_->GetUtteranceInfo().tv_usec
              <<              "\" UUID=\"" << features_->GetID() << "\" />\n" << delimiter_sym_
              << "<STARTRECOG CLIENTID=\"" << features_->GetUtteranceInfo().client_id
              <<          "\" MICINDEX=\"" << features_->GetUtteranceInfo().mic_index
              <<          "\" SOURCEID=\"" << features_->GetUtteranceInfo().source_id
              <<             "\" SUBID=\"" << features_->GetUtteranceInfo().sub_id
              <<               "\" SEC=\"" << features_->GetUtteranceInfo().tv_sec
              <<              "\" USEC=\"" << features_->GetUtteranceInfo().tv_usec
              <<              "\" UUID=\"" << features_->GetID() << "\" />\n" << delimiter_sym_;
    }
    //else if(features_->is_file_input() || features_->is_mfcnet_input()){
    else{
      ss_sock << std::fixed << std::setprecision(6)
              << "<SOURCEINFO SOURCEID=\"" << features_->GetSourceInfo().source_id
              <<           "\" AZIMUTH=\"" << features_->GetSourceInfo().azimuth
              <<         "\" ELEVATION=\"" << features_->GetSourceInfo().elevation
              <<               "\" SEC=\"" << features_->GetSourceInfo().tv_sec
              <<              "\" USEC=\"" << features_->GetSourceInfo().tv_usec << "\" />\n" << delimiter_sym_
              << "<STARTRECOG SOURCEID=\"" << features_->GetSourceInfo().source_id << "\" />\n" << delimiter_sym_;
    }
    result_client_->broadcast(ss_sock.str());
  }

  //DEBUG//KALDI_LOG << "started recognize key id=" << id_;
  Timer timer; timer.Reset();
  //decoder_.Decode(decodable_);
  decoder_.InitDecoding();

  int progress_iframes = 0;
  //int progress_dframes = progress_output_interval_;
  int progress_dframes = 0;
  int progress_wait = static_cast<int>(progress_output_interval_ * 0.5f);
  while(true){
    if(features_){
      //DEBUG//KALDI_LOG << "receiving features key id=" << id_;
      features_->Receive();
    }
    if(silence_weighting->Active() && feature_pipeline->IvectorFeature() != NULL){
      silence_weighting->ComputeCurrentTraceback(decoder_);
      silence_weighting->GetDeltaWeights(feature_pipeline->NumFramesReady(), &delta_weights);
      feature_pipeline->IvectorFeature()->UpdateFrameWeights(delta_weights);
    }
    if(features_if_->IsLastFrame(features_if_->NumFramesReady() - 1)){
      decoder_.AdvanceDecoding(decodable_);
      //DEBUG//KALDI_LOG << "final frame recognized key id=" << id_;
      break;
    }
    if(progress_output_interval_ > 0){
      bool not_enough = true;
      while(features_if_->NumFramesReady() > progress_iframes){
        not_enough = false;
        decoder_.AdvanceDecoding(decodable_, progress_output_interval_);
        progress_iframes += progress_output_interval_;
        if(decoder_.NumFramesDecoded() > progress_dframes){
          progress_dframes = decoder_.NumFramesDecoded();
          WriteIntermediateResult(progress_dframes);
        }
      }
      if(not_enough){
        multiplatform::time::ms_sleep(progress_wait);
      }
    }
    else{
      multiplatform::time::ms_sleep(500);
    }
  }

  decoder_.FinalizeDecoding();
  //DEBUG//KALDI_LOG << "finalized decoding key id=" << id_;
  double elapsed = timer.Elapsed();
  KALDI_LOG << elapsed << " sec, RT=" << (elapsed * 100.0 / features_if_->NumFramesReady());

  if(features_){
    std::ostringstream ss_sock;
    if(features_->is_ex_mfcnet_input()){
      ss_sock << "<ENDRECOG CLIENTID=\"" << features_->GetUtteranceInfo().client_id
              <<        "\" MICINDEX=\"" << features_->GetUtteranceInfo().mic_index
              <<        "\" SOURCEID=\"" << features_->GetUtteranceInfo().source_id
              <<           "\" SUBID=\"" << features_->GetUtteranceInfo().sub_id
              <<             "\" SEC=\"" << features_->GetUtteranceInfo().tv_sec
              <<            "\" USEC=\"" << features_->GetUtteranceInfo().tv_usec
              <<            "\" UUID=\"" << features_->GetID() << "\" />\n" << delimiter_sym_;
    }
    //else if(features_->is_file_input() || features_->is_mfcnet_input()){
    else{
      ss_sock << "<ENDRECOG SOURCEID=\"" << features_->GetSourceInfo().source_id << "\" />\n" << delimiter_sym_;
    }
    result_client_->broadcast(ss_sock.str());
  }
  WriteResult();

  if(enable_lat_combine_){
    if(features_){
      timer.Reset();
      cmb_decoder_.InitDecoding();
      cmb_decoder_.AdvanceDecoding(cmb_decodable_);
      cmb_decoder_.FinalizeDecoding();
      elapsed = timer.Elapsed();
      KALDI_LOG << elapsed << " sec, RT=" << (elapsed * 100.0 / features_if_->NumFramesReady());
    }
    WriteRescoringResult();
  }

  if(features_){
    result_client_->add_recogend_key(id_);
  }
}

void NnetRecognizer::WriteIntermediateResult(kaldi::int32 frame)
{
  if(word_syms_ == NULL || decoder_.NumFramesDecoded() <= 0){
    std::ostringstream ss_sock;
    if(features_->is_ex_mfcnet_input()){
      ss_sock << "<RECOGFAIL CLIENTID=\"" << features_->GetUtteranceInfo().client_id
              <<         "\" MICINDEX=\"" << features_->GetUtteranceInfo().mic_index
              <<         "\" SOURCEID=\"" << features_->GetUtteranceInfo().source_id
              <<            "\" SUBID=\"" << features_->GetUtteranceInfo().sub_id
              <<              "\" SEC=\"" << features_->GetUtteranceInfo().tv_sec
              <<             "\" USEC=\"" << features_->GetUtteranceInfo().tv_usec
              <<             "\" UUID=\"" << features_->GetID() << "\" />\n" << delimiter_sym_;
    }
    //else if(features_->is_file_input() || features_->is_mfcnet_input()){
    else{
      ss_sock << "<RECOGFAIL SOURCEID=\"" << features_->GetSourceInfo().source_id << "\"/>\n" << delimiter_sym_;
    }
    result_client_->broadcast(ss_sock.str());
    return;
  }

  Lattice best_path_lat;
  decoder_.GetBestPath(&best_path_lat, false);
  std::vector<int32> phones_aligned;
  std::vector<int32> words;
  LatticeWeight weight;

  GetPhoneAlignmentSequence(best_path_lat, &phones_aligned);
  GetWordAlignmentSequence(best_path_lat, trans_model_, lexicon_info_, &words, &weight);
  // implementing now
  //GetWordWithPhoneAlignmentSequence(best_path_lat, trans_model_, lexicon_info_, &words, &phones_aligned &weight);

  {
    std::ostringstream ss_sock;
    if(features_->is_ex_mfcnet_input()){
      ss_sock << "<RECOGOUT CLIENTID=\"" << features_->GetUtteranceInfo().client_id
              <<        "\" MICINDEX=\"" << features_->GetUtteranceInfo().mic_index
              <<        "\" SOURCEID=\"" << features_->GetUtteranceInfo().source_id
              <<           "\" SUBID=\"" << features_->GetUtteranceInfo().sub_id
              <<             "\" SEC=\"" << features_->GetUtteranceInfo().tv_sec
              <<            "\" USEC=\"" << features_->GetUtteranceInfo().tv_usec
              <<            "\" UUID=\"" << features_->GetID() << "\"";
    }
    //else if(features_->is_file_input() || features_->is_mfcnet_input()){
    else{
      ss_sock << "<RECOGOUT SOURCEID=\"" << features_->GetSourceInfo().source_id << "\"";
    }
    if(lm_name_ != ""){ ss_sock << " NAME=\"" << lm_name_ << "\""; }
    ss_sock << std::fixed << std::setprecision(6)
            << ">\n"
            << " <PHYPO RANK=\"" << 1
            <<     "\" SCORE=\"" << -(double)(weight.Value1() + weight.Value2())
            <<   "\" AMSCORE=\"" << -(double)weight.Value2()
            <<   "\" LMSCORE=\"" << -(double)weight.Value1()
            <<     "\" FRAME=\"" << frame
            <<      "\" TIME=\"" << time(NULL)
            << "\">\n";
    size_t i = 0;
    //int word_index = 0;
    while(i < words.size()){
      std::string classid = word_syms_->Find(words[i]);
      std::string word = string_tokenizer(classid, "+")[0];
      //std::string word = (word_index < string_tokenizer(classid, "+").size()) ? string_tokenizer(classid, "+")[word_index] : string_tokenizer(classid, "+")[0];
      if(classid == ""){ KALDI_ERR << "Word-id " << words[i] << " not in symbol table."; }
      ss_sock << std::fixed << std::setprecision(3)
              << "  <WHYPO WORD=\"" << word
              <<    "\" CLASSID=\"" << classid
              <<      "\" PHONE=\"" << ""
              <<         "\" CM=\"" << 1.0f << "\"/>\n";
      i++;
    }
    ss_sock << " </PHYPO>\n</RECOGOUT>\n" << delimiter_sym_;

    result_client_->broadcast(ss_sock.str());
  }
}

void NnetRecognizer::WriteResult(std::ostream& out)
{
  if(word_syms_ == NULL || decoder_.NumFramesDecoded() <= 0){
    // Write error result for socket
    if(features_){
      std::ostringstream ss_sock;
      if(features_->is_ex_mfcnet_input()){
        ss_sock << "<RECOGFAIL CLIENTID=\"" << features_->GetUtteranceInfo().client_id
                <<         "\" MICINDEX=\"" << features_->GetUtteranceInfo().mic_index
                <<         "\" SOURCEID=\"" << features_->GetUtteranceInfo().source_id
                <<            "\" SUBID=\"" << features_->GetUtteranceInfo().sub_id
                <<              "\" SEC=\"" << features_->GetUtteranceInfo().tv_sec
                <<             "\" USEC=\"" << features_->GetUtteranceInfo().tv_usec
                <<             "\" UUID=\"" << features_->GetID() << "\" />\n" << delimiter_sym_;
      }
      //else if(features_->is_file_input() || features_->is_mfcnet_input()){
      else{
        ss_sock << "<RECOGFAIL SOURCEID=\"" << features_->GetSourceInfo().source_id << "\"/>\n" << delimiter_sym_;
      }
      result_client_->broadcast(ss_sock.str());
    }

    // Write error result for stdout
    if(enable_print_result_ && enable_print_sentence_){
      std::stringstream ss_cout;
      if(emulate_julius_){
        if(features_){
          if(features_->is_ex_mfcnet_input()){
            ss_cout << std::fixed << std::setprecision(6)
                    <<   "client_id = " << features_->GetUtteranceInfo().client_id
                    << ", mic_index = " << features_->GetUtteranceInfo().mic_index
                    << ", source_id = " << features_->GetUtteranceInfo().source_id
                    <<    ", sub_id = " << features_->GetUtteranceInfo().sub_id
                    <<   ", azimuth = " << features_->GetUtteranceInfo().azimuth
                    << ", elevation = " << features_->GetUtteranceInfo().elevation
                    <<       ", sec = " << features_->GetUtteranceInfo().tv_sec
                    <<      ", usec = " << features_->GetUtteranceInfo().tv_usec
                    <<      ", uuid = " << features_->GetID() << "\n";
          }
          //else if(features_->is_file_input() || features_->is_mfcnet_input()){
          else{
            ss_cout << std::fixed << std::setprecision(6)
                    <<   "source_id = " << features_->GetSourceInfo().source_id
                    <<   ", azimuth = " << features_->GetSourceInfo().azimuth
                    << ", elevation = " << features_->GetSourceInfo().elevation
                    <<       ", sec = " << features_->GetSourceInfo().tv_sec
                    <<      ", usec = " << features_->GetSourceInfo().tv_usec << "\n";
          }
        }
        else{
          ss_cout << "source_id = -1, azimuth = 0.000000, elevation = 0.000000, sec = 0, usec = 0 \n";
        }
        ss_cout << "### Recognition: 2nd pass (RL heuristic best-first) \n"
                << "STAT: 00 \n"
                << "sentence1: \n"
                << "wseq1: \n"
                << "phseq1: \n"
                << "cmscore1: \n"
                << "score1: \n\n"
                << "ERROR: recognition failed.";
      }
      else{
        ss_cout << "recognition failed\n"
                << "result: \n"
                << "phone_seq: ";
      }
      out << ss_cout.str() << std::endl;
    }

    return;
  }

  if(nbest_count_ <= 1){
    Lattice best_path_lat;
    decoder_.GetBestPath(&best_path_lat);
    std::vector<int32> phones_aligned;
    std::vector<int32> words;
    LatticeWeight weight;

    GetPhoneAlignmentSequence(best_path_lat, &phones_aligned);
    GetWordAlignmentSequence(best_path_lat, trans_model_, lexicon_info_, &words, &weight);
    // implementing now
    //GetWordWithPhoneAlignmentSequence(best_path_lat, trans_model_, lexicon_info_, &words, &phones_aligned &weight);

    std::vector<std::string> str_classids; str_classids.resize(words.size());
    std::vector<std::string> str_words; str_words.resize(words.size());

    if(word_syms_ != NULL){
      size_t i = 0;
      //int word_index = 0;
      while(i < words.size()){
        str_classids[i] = word_syms_->Find(words[i]);
        str_words[i] = string_tokenizer(str_classids[i], "+")[0];
        //str_words[i] = (word_index < string_tokenizer(str_classids[i], "+").size()) ? string_tokenizer(str_classids[i], "+")[word_index] : string_tokenizer(str_classids[i], "+")[0];
        if(str_classids[i] == ""){ KALDI_ERR << "Word-id " << words[i] << " not in symbol table."; }
        i++;
      }
    }
    else{
      size_t i = 0;
      while(i < words.size()){
        str_classids[i] = str_words[i] = std::to_string(words[i]);
        i++;
      }
    }

    // Write result for socket
    if(features_){
      std::ostringstream ss_sock;

      if(features_->is_ex_mfcnet_input()){
        ss_sock << "<RECOGOUT CLIENTID=\"" << features_->GetUtteranceInfo().client_id
                <<        "\" MICINDEX=\"" << features_->GetUtteranceInfo().mic_index
                <<        "\" SOURCEID=\"" << features_->GetUtteranceInfo().source_id
                <<           "\" SUBID=\"" << features_->GetUtteranceInfo().sub_id
                <<             "\" SEC=\"" << features_->GetUtteranceInfo().tv_sec
                <<            "\" USEC=\"" << features_->GetUtteranceInfo().tv_usec
                <<            "\" UUID=\"" << features_->GetID() << "\"";
      }
      //else if(features_->is_file_input() || features_->is_mfcnet_input()){
      else{
        ss_sock << "<RECOGOUT SOURCEID=\"" << features_->GetSourceInfo().source_id << "\"";
      }
      if(lm_name_ != ""){ ss_sock << " NAME=\"" << lm_name_ << "\""; }
      ss_sock << std::fixed << std::setprecision(6)
              << ">\n"
              << " <SHYPO RANK=\"" << 1
              <<     "\" SCORE=\"" << -(double)(weight.Value1() + weight.Value2())
              <<   "\" AMSCORE=\"" << -(double)weight.Value2()
              <<   "\" LMSCORE=\"" << -(double)weight.Value1() << "\">\n";

      size_t i = 0;
      while(i < words.size()){
        ss_sock << std::fixed << std::setprecision(3)
                << "  <WHYPO WORD=\"" << str_words[i]
                <<    "\" CLASSID=\"" << str_classids[i]
                <<      "\" PHONE=\"" << ""
                <<         "\" CM=\"" << 1.0f << "\"/>\n";
        i++;
      }
      ss_sock << " </SHYPO>\n</RECOGOUT>\n" << delimiter_sym_;

      result_client_->broadcast(ss_sock.str());
    }

    // Write result for stdout
    {
      if(enable_print_result_ && enable_print_sentence_){
        std::stringstream ss_cout;

        if(emulate_julius_){
          if(features_){
            if(features_->is_ex_mfcnet_input()){
              ss_cout << std::fixed << std::setprecision(6)
                      <<   "client_id = " << features_->GetUtteranceInfo().client_id
                      << ", mic_index = " << features_->GetUtteranceInfo().mic_index
                      << ", source_id = " << features_->GetUtteranceInfo().source_id
                      <<    ", sub_id = " << features_->GetUtteranceInfo().sub_id
                      <<   ", azimuth = " << features_->GetUtteranceInfo().azimuth
                      << ", elevation = " << features_->GetUtteranceInfo().elevation
                      <<       ", sec = " << features_->GetUtteranceInfo().tv_sec
                      <<      ", usec = " << features_->GetUtteranceInfo().tv_usec
                      <<      ", uuid = " << features_->GetID() << "\n";
            }
            //else if(features_->is_file_input() || features_->is_mfcnet_input()){
            else{
              ss_cout << std::fixed << std::setprecision(6)
                      <<   "source_id = " << features_->GetSourceInfo().source_id
                      <<   ", azimuth = " << features_->GetSourceInfo().azimuth
                      << ", elevation = " << features_->GetSourceInfo().elevation
                      <<       ", sec = " << features_->GetSourceInfo().tv_sec
                      <<      ", usec = " << features_->GetSourceInfo().tv_usec << "\n";
            }
          }
          else{
            ss_cout << "source_id = -1, azimuth = 0.000000, elevation = 0.000000, sec = 0, usec = 0 \n";
          }

          ss_cout << "### Recognition: 2nd pass (RL heuristic best-first) \n"
                  << "STAT: 00 \n"
                  << "sentence1: ";
          {
            size_t i = 0;
            while(i < words.size()){ ss_cout << str_words[i++] << " "; }
          }
          ss_cout << "\n";

          ss_cout << "wseq1: ";
          {
            size_t i = 0;
            while(i < words.size()){ ss_cout << str_classids[i++] << " "; }
          }
          ss_cout << "\n";

          ss_cout << "phseq1: ";
          if(phone_syms_ != NULL){
            size_t i = 0;
            while(i < phones_aligned.size()){
              kaldi::int32 tstate = trans_model_.TransitionIdToTransitionState(phones_aligned[i]);
              kaldi::int32 phone = trans_model_.TransitionStateToPhone(tstate);
              std::string s = phone_syms_->Find(phone);
              if(s == ""){ KALDI_ERR << "Phone-id " << phone << " not in symbol table."; }
              ss_cout << s << " ";
              i++;
            }
          }
          ss_cout << "\n";

          ss_cout << "cmscore1: ";
          {
            size_t i = 0;
            while(i < words.size()){
              ss_cout << std::fixed << std::setprecision(3) << 1.0f << " ";
              i++;
            }
          }
          ss_cout << "\n";

          ss_cout << std::fixed << std::setprecision(6)
                  << "score1: " << -(double)(weight.Value1() + weight.Value2())
                  << " ( AM: " << -(double)weight.Value2() << ", LM: " << -(double)weight.Value1() << " )";
        }
        else{
          ss_cout << "result: ";
          {
            size_t i = 0;
            while(i < words.size()){ ss_cout << str_classids[i++] << " "; }
          }
          ss_cout << "\n";

          ss_cout << "phone_seq: ";
          if(phone_syms_ != NULL){
            size_t i = 0;
            while(i < phones_aligned.size()){
              kaldi::int32 tstate = trans_model_.TransitionIdToTransitionState(phones_aligned[i]);
              kaldi::int32 phone = trans_model_.TransitionStateToPhone(tstate);
              std::string s = phone_syms_->Find(phone);
              if(s == ""){ KALDI_ERR << "Phone-id " << phone << " not in symbol table."; }
              ss_cout << s << " ";
              i++;
            }
          }
          ss_cout << "\n";

          ss_cout << std::fixed << std::setprecision(6)
                  <<     "score=" << -(double)(weight.Value1() + weight.Value2())
                  << ", amscore=" << -(double)weight.Value2()
                  << ", lmscore=" << -(double)weight.Value1();
        }

        out << ss_cout.str() << std::endl;
      }
    }
  }
  else{
    kaldi::Lattice nbest_lat;

    kaldi::WordAlignLatticeLexiconOpts lexicon_opts;

    decoder_.GetRawLattice(&nbest_lat);

    {
      std::vector<kaldi::Lattice> nbest_lats;
      {
        kaldi::Lattice nbest_lat_;
        fst::ShortestPath(nbest_lat, &nbest_lat_, nbest_count_);
        fst::ConvertNbestToVector(nbest_lat_, &nbest_lats);
      }

      std::ostringstream ss_sock;
      std::stringstream ss_cout;

      // Write result for socket
      if(features_){
        if(features_->is_ex_mfcnet_input()){
          ss_sock << "<RECOGOUT CLIENTID=\"" << features_->GetUtteranceInfo().client_id
                  <<        "\" MICINDEX=\"" << features_->GetUtteranceInfo().mic_index
                  <<        "\" SOURCEID=\"" << features_->GetUtteranceInfo().source_id
                  <<           "\" SUBID=\"" << features_->GetUtteranceInfo().sub_id
                  <<             "\" SEC=\"" << features_->GetUtteranceInfo().tv_sec
                  <<            "\" USEC=\"" << features_->GetUtteranceInfo().tv_usec
                  <<            "\" UUID=\"" << features_->GetID() << "\"";
        }
        //else if(features_->is_file_input() || features_->is_mfcnet_input()){
        else{
          ss_sock << "<RECOGOUT SOURCEID=\"" << features_->GetSourceInfo().source_id << "\"";
        }
        if(lm_name_ != ""){ ss_sock << " NAME=\"" << lm_name_ << "\""; }
        ss_sock << ">\n";
      }

      // Write result for stdout
      {
        if(enable_print_result_ && enable_print_sentence_){
          if(emulate_julius_){
            if(features_){
              if(features_->is_ex_mfcnet_input()){
                ss_cout << std::fixed << std::setprecision(6)
                        <<   "client_id = " << features_->GetUtteranceInfo().client_id
                        << ", mic_index = " << features_->GetUtteranceInfo().mic_index
                        << ", source_id = " << features_->GetUtteranceInfo().source_id
                        <<    ", sub_id = " << features_->GetUtteranceInfo().sub_id
                        <<   ", azimuth = " << features_->GetUtteranceInfo().azimuth
                        << ", elevation = " << features_->GetUtteranceInfo().elevation
                        <<       ", sec = " << features_->GetUtteranceInfo().tv_sec
                        <<      ", usec = " << features_->GetUtteranceInfo().tv_usec
                        <<      ", uuid = " << features_->GetID() << "\n";
              }
              //else if(features_->is_file_input() || features_->is_mfcnet_input()){
              else{
                ss_cout << std::fixed << std::setprecision(6)
                        <<   "source_id = " << features_->GetSourceInfo().source_id
                        <<   ", azimuth = " << features_->GetSourceInfo().azimuth
                        << ", elevation = " << features_->GetSourceInfo().elevation
                        <<       ", sec = " << features_->GetSourceInfo().tv_sec
                        <<      ", usec = " << features_->GetSourceInfo().tv_usec << "\n";
              }
            }

            ss_cout << "### Recognition: 2nd pass (RL heuristic best-first) \n"
                    << "STAT: 00 \n";
          }
        }
      }

      for(kaldi::int32 k = 0; k < static_cast<kaldi::int32>(nbest_lats.size()); k++){
        //std::cout << "[CompactLattice1:Rank=" << k << "]" << std::endl;
        fst::ScaleLattice(fst::LatticeScale(opt_.acoustic_scale, 1.0/opt_.acoustic_scale), &(nbest_lats[k]));
        {
          std::vector<int32> nbest_phones_aligned;
          std::vector<int32> nbest_words;
          LatticeWeight nbest_weight;


          kaldi::CompactLattice nbest_clat;
          kaldi::CompactLattice nbest_clat_;
          kaldi::Lattice nbest_lat_;
          std::vector<int32> nbest_alignment;
          fst::ConvertLattice(nbest_lats[k], &nbest_clat);
          fst::RemoveEpsLocal(&nbest_clat);
          fst::Determinize(nbest_clat, &nbest_clat_);
          kaldi::WordAlignLatticeLexicon(nbest_clat_, trans_model_, lexicon_info_, lexicon_opts, &nbest_clat);

          kaldi::CompactLatticeShortestPath(nbest_clat, &nbest_clat_);
          fst::ConvertLattice(nbest_clat_, &nbest_lat_);
          fst::GetLinearSymbolSequence(nbest_lat_, &nbest_alignment, &nbest_words, &nbest_weight);

          GetPhoneAlignmentSequence(nbest_lat_, &nbest_phones_aligned);

          {
            std::vector<std::string> str_classids; str_classids.resize(nbest_words.size());
            std::vector<std::string> str_words; str_words.resize(nbest_words.size());

            if(word_syms_ != NULL){
              size_t i = 0;
              //int word_index = 0;
              while(i < nbest_words.size()){
                str_classids[i] = word_syms_->Find(nbest_words[i]);
                str_words[i] = string_tokenizer(str_classids[i], "+")[0];
                //str_words[i] = (word_index < string_tokenizer(str_classids[i], "+").size()) ? string_tokenizer(str_classids[i], "+")[word_index] : string_tokenizer(str_classids[i], "+")[0];
                if(str_classids[i] == ""){ KALDI_ERR << "Word-id " << nbest_words[i] << " not in symbol table."; }
                i++;
              }
            }
            else{
              size_t i = 0;
              while(i < nbest_words.size()){
                str_classids[i] = str_words[i] = std::to_string(nbest_words[i]);
                i++;
              }
            }

            // Write result for socket
            if(features_){
              ss_sock << std::fixed << std::setprecision(6)
                      << " <SHYPO RANK=\"" << k + 1
                      <<     "\" SCORE=\"" << -(double)(nbest_weight.Value1() + nbest_weight.Value2())
                      <<   "\" AMSCORE=\"" << -(double)nbest_weight.Value2()
                      <<   "\" LMSCORE=\"" << -(double)nbest_weight.Value1() << "\">\n";

              size_t i = 0;
              while(i < nbest_words.size()){
                ss_sock << std::fixed << std::setprecision(3)
                        << "  <WHYPO WORD=\"" << str_words[i]
                        <<    "\" CLASSID=\"" << str_classids[i]
                        <<      "\" PHONE=\"" << ""
                        <<         "\" CM=\"" << 1.0f << "\"/>\n";
                i++;
              }
              ss_sock << " </SHYPO>\n";
            }

            // Write result for stdout
            {
              if(enable_print_result_ && enable_print_sentence_){
                if(emulate_julius_){
                  ss_cout << "sentence" << k + 1 << ": ";
                  {
                    size_t i = 0;
                    while(i < nbest_words.size()){ ss_cout << str_words[i++] << " "; }
                  }
                  ss_cout << "\n";

                  ss_cout << "wseq" << k + 1 << ": ";
                  {
                    size_t i = 0;
                    while(i < nbest_words.size()){ ss_cout << str_classids[i++] << " "; }
                  }
                  ss_cout << "\n";

                  ss_cout << "phseq" << k + 1 << ": ";
                  if(phone_syms_ != NULL){
                    size_t i = 0;
                    while(i < nbest_phones_aligned.size()){
                      kaldi::int32 tstate = trans_model_.TransitionIdToTransitionState(nbest_phones_aligned[i]);
                      kaldi::int32 phone = trans_model_.TransitionStateToPhone(tstate);
                      std::string s = phone_syms_->Find(phone);
                      if(s == ""){ KALDI_ERR << "Phone-id " << phone << " not in symbol table."; }
                      ss_cout << s << " ";
                      i++;
                    }
                  }
                  ss_cout << "\n";

                  ss_cout << "cmscore" << k + 1 << ": ";
                  {
                    size_t i = 0;
                    while(i < nbest_words.size()){
                      ss_cout << std::fixed << std::setprecision(3) << 1.0f << " ";
                      i++;
                    }
                  }
                  ss_cout << "\n";

                  ss_cout << std::fixed << std::setprecision(6)
                          << "score" << k + 1 << ": " << -(double)(nbest_weight.Value1() + nbest_weight.Value2())
                          << " ( AM: " << -(double)nbest_weight.Value2() << ", LM: " << -(double)nbest_weight.Value1() << " )\n";
                }
                else{
                  ss_cout << "result[" << k + 1 << "]: ";
                  {
                    size_t i = 0;
                    while(i < nbest_words.size()){ ss_cout << str_classids[i++] << " "; }
                  }
                  ss_cout << "\n";

                  ss_cout << "phone_seq[" << k + 1 << "]: ";
                  if(phone_syms_ != NULL){
                    size_t i = 0;
                    while(i < nbest_phones_aligned.size()){
                      kaldi::int32 tstate = trans_model_.TransitionIdToTransitionState(nbest_phones_aligned[i]);
                      kaldi::int32 phone = trans_model_.TransitionStateToPhone(tstate);
                      std::string s = phone_syms_->Find(phone);
                      if(s == ""){ KALDI_ERR << "Phone-id " << phone << " not in symbol table."; }
                      ss_cout << s << " ";
                      i++;
                    }
                  }
                  ss_cout << "\n";

                  ss_cout << std::fixed << std::setprecision(6)
                          <<     "score[" << k + 1 << "]=" << -(double)(nbest_weight.Value1() + nbest_weight.Value2())
                          << ", amscore[" << k + 1 << "]=" << -(double)nbest_weight.Value2()
                          << ", lmscore[" << k + 1 << "]=" << -(double)nbest_weight.Value1()
                          << "\n";
                }
              }
            }
          }
        }
      }

      // Write result for socket
      if(features_){
        ss_sock << "</RECOGOUT>\n" << delimiter_sym_;

        result_client_->broadcast(ss_sock.str());
      }

      // Write result for stdout
      {
        out << ss_cout.str() << std::endl;
      }
    }
  }
}

void NnetRecognizer::WriteRescoringResult(std::ostream& out)
{
  if(word_syms_ == NULL || decoder_.NumFramesDecoded() <= 0 || cmb_decoder_.NumFramesDecoded() <= 0){
    // Write error result for socket
    if(features_){
      std::ostringstream ss_sock;
      if(features_->is_ex_mfcnet_input()){
        ss_sock << "<RECOGFAIL CLIENTID=\"" << features_->GetUtteranceInfo().client_id
                <<         "\" MICINDEX=\"" << features_->GetUtteranceInfo().mic_index
                <<         "\" SOURCEID=\"" << features_->GetUtteranceInfo().source_id
                <<            "\" SUBID=\"" << features_->GetUtteranceInfo().sub_id
                <<              "\" SEC=\"" << features_->GetUtteranceInfo().tv_sec
                <<             "\" USEC=\"" << features_->GetUtteranceInfo().tv_usec
                <<             "\" UUID=\"" << features_->GetID() << "\" />\n" << delimiter_sym_;
      }
      //else if(features_->is_file_input() || features_->is_mfcnet_input()){
      else{
        ss_sock << "<RECOGFAIL SOURCEID=\"" << features_->GetSourceInfo().source_id << "\"/>\n" << delimiter_sym_;
      }
      result_client_->broadcast(ss_sock.str());
    }

    // Write error result for stdout
    if(enable_print_result_ && enable_print_lat_combine_){
      std::stringstream ss_cout;
      if(emulate_julius_){
        if(features_){
          if(features_->is_ex_mfcnet_input()){
            ss_cout << std::fixed << std::setprecision(6)
                    <<   "client_id = " << features_->GetUtteranceInfo().client_id
                    << ", mic_index = " << features_->GetUtteranceInfo().mic_index
                    << ", source_id = " << features_->GetUtteranceInfo().source_id
                    <<    ", sub_id = " << features_->GetUtteranceInfo().sub_id
                    <<   ", azimuth = " << features_->GetUtteranceInfo().azimuth
                    << ", elevation = " << features_->GetUtteranceInfo().elevation
                    <<       ", sec = " << features_->GetUtteranceInfo().tv_sec
                    <<      ", usec = " << features_->GetUtteranceInfo().tv_usec
                    <<      ", uuid = " << features_->GetID() << "\n";
          }
          //else if(features_->is_file_input() || features_->is_mfcnet_input()){
          else{
            ss_cout << std::fixed << std::setprecision(6)
                    <<   "source_id = " << features_->GetSourceInfo().source_id
                    <<   ", azimuth = " << features_->GetSourceInfo().azimuth
                    << ", elevation = " << features_->GetSourceInfo().elevation
                    <<       ", sec = " << features_->GetSourceInfo().tv_sec
                    <<      ", usec = " << features_->GetSourceInfo().tv_usec << "\n";
          }
        }
        else{
          ss_cout << "source_id = -1, azimuth = 0.000000, elevation = 0.000000, sec = 0, usec = 0 \n";
        }
        ss_cout << "### Recognition: Rescoring pass \n"
                << "STAT: 00 \n"
                << "sentence1: \n"
                << "wseq1: \n"
                << "phseq1: \n"
                << "cmscore1: \n"
                << "score1: \n\n"
                << "ERROR: recognition failed.";
      }
      else{
        ss_cout << "recognition failed\n"
                << "result: \n"
                << "phone_seq: ";
      }
      out << ss_cout.str() << std::endl;
    }

    return;
  }

  ////////// CPU usage is high. (Confirmation before loading Lattice.) //////////
  if(rm_->is_cpu_warning()){
    // Write error result for socket
    if(features_){
      std::ostringstream ss_sock;
      if(features_->is_ex_mfcnet_input()){
        ss_sock << "<RECOGOUT CLIENTID=\"" << features_->GetUtteranceInfo().client_id
                <<        "\" MICINDEX=\"" << features_->GetUtteranceInfo().mic_index
                <<        "\" SOURCEID=\"" << features_->GetUtteranceInfo().source_id
                <<           "\" SUBID=\"" << features_->GetUtteranceInfo().sub_id
                <<             "\" SEC=\"" << features_->GetUtteranceInfo().tv_sec
                <<            "\" USEC=\"" << features_->GetUtteranceInfo().tv_usec
                <<            "\" UUID=\"" << features_->GetID() << "\"  TERMINATE=\"CPU usage is high.\">\n";
      }
      //else if(features_->is_file_input() || features_->is_mfcnet_input()){
      else{
        ss_sock << "<RECOGOUT SOURCEID=\"" << features_->GetSourceInfo().source_id << "\"  TERMINATE=\"CPU usage is high.\">\n";
      }
      ss_sock << "</RECOGOUT>\n" << delimiter_sym_;

      result_client_->broadcast(ss_sock.str());
    }

    // Write error result for stdout
    if(enable_print_result_ && enable_print_lat_combine_){
      std::ostringstream ss_cout;
      if(emulate_julius_){
        if(features_){
          if(features_->is_ex_mfcnet_input()){
            ss_cout << std::fixed << std::setprecision(6)
                    <<   "client_id = " << features_->GetUtteranceInfo().client_id
                    << ", mic_index = " << features_->GetUtteranceInfo().mic_index
                    << ", source_id = " << features_->GetUtteranceInfo().source_id
                    <<    ", sub_id = " << features_->GetUtteranceInfo().sub_id
                    <<   ", azimuth = " << features_->GetUtteranceInfo().azimuth
                    << ", elevation = " << features_->GetUtteranceInfo().elevation
                    <<       ", sec = " << features_->GetUtteranceInfo().tv_sec
                    <<      ", usec = " << features_->GetUtteranceInfo().tv_usec
                    <<      ", uuid = " << features_->GetID() << "\n";
          }
          //else if(features_->is_file_input() || features_->is_mfcnet_input()){
          else{
            ss_cout << std::fixed << std::setprecision(6)
                    <<   "source_id = " << features_->GetSourceInfo().source_id
                    <<   ", azimuth = " << features_->GetSourceInfo().azimuth
                    << ", elevation = " << features_->GetSourceInfo().elevation
                    <<       ", sec = " << features_->GetSourceInfo().tv_sec
                    <<      ", usec = " << features_->GetSourceInfo().tv_usec << "\n";
          }
        }
        else{
          ss_cout << "source_id = -1, azimuth = 0.000000, elevation = 0.000000, sec = 0, usec = 0 \n";
        }
        ss_cout << "### Recognition: Rescoring pass \n"
                << "STAT: 00 \n"
                << "sentence1: \n"
                << "wseq1: \n"
                << "phseq1: \n"
                << "cmscore1: \n"
                << "score1: \n\n"
                << "ERROR: CPU usage is high.";
      }
      else{
        ss_cout << "CPU usage is high\n"
                << "result: \n"
                << "phone_seq: ";
      }
      out << ss_cout.str() << std::endl;
    }

    return;
  }

  std::vector<int32> phones_aligned;
  std::vector<int32> words;
  LatticeWeight weight;

  if(enable_lat_combine_){
    kaldi::Lattice lat1;
    kaldi::Lattice lat2;
    kaldi::CompactLattice clat1;
    kaldi::CompactLattice clat2;
    kaldi::CompactLattice clat1_; // temp
    kaldi::CompactLattice clat2_; // temp
    kaldi::WordAlignLatticeLexiconOpts lexicon_opts;

    decoder_.GetRawLattice(&lat1);
    cmb_decoder_.GetRawLattice(&lat2);

    {
      std::vector<kaldi::Lattice> nbest_lats;
      {
        kaldi::Lattice nbest_lat;
        fst::ShortestPath(lat1, &nbest_lat, nbest_for_lat_combine_);
        fst::ConvertNbestToVector(nbest_lat, &nbest_lats);
      }

      if(!nbest_lats.empty()){
        //std::cout << "[CompactLattice1:Rank=" << 0 << "]" << std::endl;
        fst::ConvertLattice(nbest_lats[0], &clat1);
        fst::RemoveEpsLocal(&clat1);
        fst::Determinize(clat1, &clat1_);
        kaldi::WordAlignLatticeLexicon(clat1_, trans_model_, lexicon_info_, lexicon_opts, &clat1);
      }
      for(kaldi::int32 k = 1; k < static_cast<kaldi::int32>(nbest_lats.size()); k++){
        //std::cout << "[CompactLattice1:Rank=" << k << "]" << std::endl;
        fst::ScaleLattice(fst::LatticeScale(opt_.acoustic_scale, 1.0/opt_.acoustic_scale), &(nbest_lats[k]));
        {
          kaldi::CompactLattice nbest_clat;
          fst::ConvertLattice(nbest_lats[k], &nbest_clat);
          fst::RemoveEpsLocal(&nbest_clat);
          fst::Determinize(nbest_clat, &clat1_);
          kaldi::WordAlignLatticeLexicon(clat1_, trans_model_, lexicon_info_, lexicon_opts, &nbest_clat);
          fst::Union(&clat1, nbest_clat);
        }
      }
    }

    {
      std::vector<kaldi::Lattice> nbest_lats;
      {
        kaldi::Lattice nbest_lat;
        fst::ShortestPath(lat2, &nbest_lat, nbest_for_lat_combine_);
        fst::ConvertNbestToVector(nbest_lat, &nbest_lats);
      }

      if(!nbest_lats.empty()){
        //std::cout << "[CompactLattice2:Rank=" << 0 << "]" << std::endl;
        fst::ConvertLattice(nbest_lats[0], &clat2);
        fst::RemoveEpsLocal(&clat2);
        fst::Determinize(clat2, &clat2_);
        kaldi::WordAlignLatticeLexicon(clat2_, trans_model_, lexicon_info_, lexicon_opts, &clat2);
      }
      for(kaldi::int32 k = 1; k < static_cast<kaldi::int32>(nbest_lats.size()); k++){
        //std::cout << "[CompactLattice2:Rank=" << k << "]" << std::endl;
        fst::ScaleLattice(fst::LatticeScale(opt_.acoustic_scale, 1.0/opt_.acoustic_scale), &(nbest_lats[k]));
        {
          kaldi::CompactLattice nbest_clat;
          fst::ConvertLattice(nbest_lats[k], &nbest_clat);
          fst::RemoveEpsLocal(&nbest_clat);
          fst::Determinize(nbest_clat, &clat2_);
          kaldi::WordAlignLatticeLexicon(clat2_, trans_model_, lexicon_info_, lexicon_opts, &nbest_clat);
          fst::Union(&clat2, nbest_clat);
        }
      }
    }

    {
      float acoustic_scale = 1.0f; // temporary
      float lm_scale = 1.0f;       // temporary
      std::vector<std::vector<double> > lat_scale = fst::LatticeScale(lm_scale, acoustic_scale);
      fst::ScaleLattice(lat_scale, &clat1);
      fst::ScaleLattice(lat_scale, &clat2);
    }

    {
      bool fail = false;
      if(combine_lat_weights_.size() == 2){
        // Check lattices
        if(((&clat1)->Properties(fst::kTopSorted, false) == 0) && (fst::TopSort(&clat1) == false)){
          fail = true; // Cycles detected in clat1. Cannot normalize.
        }
        else if(((&clat2)->Properties(fst::kTopSorted, false) == 0) && (fst::TopSort(&clat2) == false)){
          fail = true; // Cycles detected in clat2. Cannot normalize.
        }
        else{}
        // Normalize lattice 1
        if(!fail){
          std::vector<double> beta;
          if(ComputeCompactLatticeBetas(clat1, &beta)){
            kaldi::CompactLattice::Arc::StateId start = (&clat1)->Start();
            kaldi::BaseFloat total_backward_cost = beta[start];
            total_backward_cost -= Log(combine_lat_weights_[0]);
            for(fst::StateIterator<CompactLattice> sit(clat1); !sit.Done(); sit.Next()){
              CompactLatticeWeight f = (&clat1)->Final(sit.Value());
              LatticeWeight w = f.Weight();
              w.SetValue1(w.Value1() + total_backward_cost);
              f.SetWeight(w);
              (&clat1)->SetFinal(sit.Value(), f);
            }
          }
          else{
            fail = true; // Failed to compute backward probabilities on clat1.
          }
        }
        // Normalize lattice 2
        if(!fail){
          std::vector<double> beta;
          if(ComputeCompactLatticeBetas(clat2, &beta)){
            kaldi::CompactLattice::Arc::StateId start = (&clat2)->Start();
            kaldi::BaseFloat total_backward_cost = beta[start];
            total_backward_cost -= Log(combine_lat_weights_[1]);
            for(fst::StateIterator<CompactLattice> sit(clat2); !sit.Done(); sit.Next()){
              CompactLatticeWeight f = (&clat2)->Final(sit.Value());
              LatticeWeight w = f.Weight();
              w.SetValue1(w.Value1() + total_backward_cost);
              f.SetWeight(w);
              (&clat2)->SetFinal(sit.Value(), f);
            }
          }
          else{
            fail = true; // Failed to compute backward probabilities on clat2.
          }
        }
      }
      else{
        fail = true; // lattice weight size must be 2.
        //// old comment // Lattice weights are positive numbers, and Lattice 1 weight + Lattice 2 weight must be 1.0f.
      }
      if(fail){
        // socket
        if(features_){
          std::ostringstream ss_sock;
          if(features_->is_ex_mfcnet_input()){
            ss_sock << "<RECOGFAIL CLIENTID=\"" << features_->GetUtteranceInfo().client_id
                    <<         "\" MICINDEX=\"" << features_->GetUtteranceInfo().mic_index
                    <<         "\" SOURCEID=\"" << features_->GetUtteranceInfo().source_id
                    <<            "\" SUBID=\"" << features_->GetUtteranceInfo().sub_id
                    <<              "\" SEC=\"" << features_->GetUtteranceInfo().tv_sec
                    <<             "\" USEC=\"" << features_->GetUtteranceInfo().tv_usec
                    <<             "\" UUID=\"" << features_->GetID() << "\" />\n" << delimiter_sym_;
          }
          //else if(features_->is_file_input() || features_->is_mfcnet_input()){
          else{
            ss_sock << "<RECOGFAIL SOURCEID=\"" << features_->GetSourceInfo().source_id << "\"/>\n" << delimiter_sym_;
          }
          result_client_->broadcast(ss_sock.str());
        }

        // stdout

        return;
      }

      ////////// CPU usage is high. (Confirmation before combining Lattice. ) //////////
      if(rm_->is_cpu_warning()){
        // Write error result for socket
        if(features_){
          std::ostringstream ss_sock;
          if(features_->is_ex_mfcnet_input()){
            ss_sock << "<RECOGOUT CLIENTID=\"" << features_->GetUtteranceInfo().client_id
                    <<        "\" MICINDEX=\"" << features_->GetUtteranceInfo().mic_index
                    <<        "\" SOURCEID=\"" << features_->GetUtteranceInfo().source_id
                    <<           "\" SUBID=\"" << features_->GetUtteranceInfo().sub_id
                    <<             "\" SEC=\"" << features_->GetUtteranceInfo().tv_sec
                    <<            "\" USEC=\"" << features_->GetUtteranceInfo().tv_usec
                    <<            "\" UUID=\"" << features_->GetID() << "\"  TERMINATE=\"CPU usage is high.\">\n";
          }
          //else if(features_->is_file_input() || features_->is_mfcnet_input()){
          else{
            ss_sock << "<RECOGOUT SOURCEID=\"" << features_->GetSourceInfo().source_id << "\"  TERMINATE=\"CPU usage is high.\">\n";
          }
          ss_sock << "</RECOGOUT>\n" << delimiter_sym_;

          result_client_->broadcast(ss_sock.str());
        }

        // Write error result for stdout
        if(enable_print_result_ && enable_print_lat_combine_){
          std::ostringstream ss_cout;
          if(emulate_julius_){
            if(features_){
              if(features_->is_ex_mfcnet_input()){
                ss_cout << std::fixed << std::setprecision(6)
                        <<   "client_id = " << features_->GetUtteranceInfo().client_id
                        << ", mic_index = " << features_->GetUtteranceInfo().mic_index
                        << ", source_id = " << features_->GetUtteranceInfo().source_id
                        <<    ", sub_id = " << features_->GetUtteranceInfo().sub_id
                        <<   ", azimuth = " << features_->GetUtteranceInfo().azimuth
                        << ", elevation = " << features_->GetUtteranceInfo().elevation
                        <<       ", sec = " << features_->GetUtteranceInfo().tv_sec
                        <<      ", usec = " << features_->GetUtteranceInfo().tv_usec
                        <<      ", uuid = " << features_->GetID() << "\n";
              }
              //else if(features_->is_file_input() || features_->is_mfcnet_input()){
              else{
                ss_cout << std::fixed << std::setprecision(6)
                        <<   "source_id = " << features_->GetSourceInfo().source_id
                        <<   ", azimuth = " << features_->GetSourceInfo().azimuth
                        << ", elevation = " << features_->GetSourceInfo().elevation
                        <<       ", sec = " << features_->GetSourceInfo().tv_sec
                        <<      ", usec = " << features_->GetSourceInfo().tv_usec << "\n";
              }
            }
            else{
              ss_cout << "source_id = -1, azimuth = 0.000000, elevation = 0.000000, sec = 0, usec = 0 \n";
            }
            ss_cout << "### Recognition: Rescoring pass \n"
                    << "STAT: 00 \n"
                    << "sentence1: \n"
                    << "wseq1: \n"
                    << "phseq1: \n"
                    << "cmscore1: \n"
                    << "score1: \n\n"
                    << "ERROR: CPU usage is high.";
          }
          else{
            ss_cout << "CPU usage is high\n"
                    << "result: \n"
                    << "phone_seq: ";
          }
          out << ss_cout.str() << std::endl;
        }

        return;
      }

      fst::Union(&clat1, clat2);
    }

    ////////// CPU usage is high. (Confirmation before decoding. ) //////////
    if(rm_->is_cpu_warning()){
      // Write error result for socket
      if(features_){
        std::ostringstream ss_sock;
        if(features_->is_ex_mfcnet_input()){
          ss_sock << "<RECOGOUT CLIENTID=\"" << features_->GetUtteranceInfo().client_id
                  <<        "\" MICINDEX=\"" << features_->GetUtteranceInfo().mic_index
                  <<        "\" SOURCEID=\"" << features_->GetUtteranceInfo().source_id
                  <<           "\" SUBID=\"" << features_->GetUtteranceInfo().sub_id
                  <<             "\" SEC=\"" << features_->GetUtteranceInfo().tv_sec
                  <<            "\" USEC=\"" << features_->GetUtteranceInfo().tv_usec
                  <<            "\" UUID=\"" << features_->GetID() << "\"  TERMINATE=\"CPU usage is high.\">\n";
        }
        //else if(features_->is_file_input() || features_->is_mfcnet_input()){
        else{
          ss_sock << "<RECOGOUT SOURCEID=\"" << features_->GetSourceInfo().source_id << "\"  TERMINATE=\"CPU usage is high.\">\n";
        }
        ss_sock << "</RECOGOUT>\n" << delimiter_sym_;

        result_client_->broadcast(ss_sock.str());
      }

      // Write error result for stdout
      if(enable_print_result_ && enable_print_lat_combine_){
        std::ostringstream ss_cout;
        if(emulate_julius_){
          if(features_){
            if(features_->is_ex_mfcnet_input()){
              ss_cout << std::fixed << std::setprecision(6)
                      <<   "client_id = " << features_->GetUtteranceInfo().client_id
                      << ", mic_index = " << features_->GetUtteranceInfo().mic_index
                      << ", source_id = " << features_->GetUtteranceInfo().source_id
                      <<    ", sub_id = " << features_->GetUtteranceInfo().sub_id
                      <<   ", azimuth = " << features_->GetUtteranceInfo().azimuth
                      << ", elevation = " << features_->GetUtteranceInfo().elevation
                      <<       ", sec = " << features_->GetUtteranceInfo().tv_sec
                      <<      ", usec = " << features_->GetUtteranceInfo().tv_usec
                      <<      ", uuid = " << features_->GetID() << "\n";
            }
            //else if(features_->is_file_input() || features_->is_mfcnet_input()){
            else{
              ss_cout << std::fixed << std::setprecision(6)
                      <<   "source_id = " << features_->GetSourceInfo().source_id
                      <<   ", azimuth = " << features_->GetSourceInfo().azimuth
                      << ", elevation = " << features_->GetSourceInfo().elevation
                      <<       ", sec = " << features_->GetSourceInfo().tv_sec
                      <<      ", usec = " << features_->GetSourceInfo().tv_usec << "\n";
            }
          }
          else{
            ss_cout << "source_id = -1, azimuth = 0.000000, elevation = 0.000000, sec = 0, usec = 0 \n";
          }
          ss_cout << "### Recognition: Rescoring pass \n"
                  << "STAT: 00 \n"
                  << "sentence1: \n"
                  << "wseq1: \n"
                  << "phseq1: \n"
                  << "cmscore1: \n"
                  << "score1: \n\n"
                  << "ERROR: CPU usage is high.";
        }
        else{
          ss_cout << "CPU usage is high\n"
                  << "result: \n"
                  << "phone_seq: ";
        }
        out << ss_cout.str() << std::endl;
      }

      return;
    }

    kaldi::MinimumBayesRisk mbr(clat1);
    words = mbr.GetOneBest();
    weight.SetValue1(-0.0f);
    weight.SetValue2(-0.0f);
  }
  else if(false){ // This mode just outputs the offline results of the second model.
    kaldi::Lattice best_path_lat;
    cmb_decoder_.GetBestPath(&best_path_lat);
    GetPhoneAlignmentSequence(best_path_lat, &phones_aligned);
    GetWordAlignmentSequence(best_path_lat, cmb_trans_model_, cmb_lexicon_info_, &words, &weight);
  }
  else{
    return;
  }

  ////////// CPU usage is high. //////////
  ////////// Since the decoding is finished, We will not stop recognition result output.
  // if(rm_->is_cpu_warning()){ }

  {
    std::vector<std::string> str_classids; str_classids.resize(words.size());
    std::vector<std::string> str_words; str_words.resize(words.size());

    if(word_syms_ != NULL){
      size_t i = 0;
      //int word_index = 0;
      while(i < words.size()){
        str_classids[i] = word_syms_->Find(words[i]);
        str_words[i] = string_tokenizer(str_classids[i], "+")[0];
        //str_words[i] = (word_index < string_tokenizer(str_classids[i], "+").size()) ? string_tokenizer(str_classids[i], "+")[word_index] : string_tokenizer(str_classids[i], "+")[0];
        if(str_classids[i] == ""){ KALDI_ERR << "Word-id " << words[i] << " not in symbol table."; }
        i++;
      }
    }
    else{
      size_t i = 0;
      while(i < words.size()){
        str_classids[i] = str_words[i] = std::to_string(words[i]);
        i++;
      }
    }

    // Write result for socket
    if(features_){
      std::ostringstream ss_sock;

      if(features_->is_ex_mfcnet_input()){
        ss_sock << "<RECOGOUT CLIENTID=\"" << features_->GetUtteranceInfo().client_id
                <<        "\" MICINDEX=\"" << features_->GetUtteranceInfo().mic_index
                <<        "\" SOURCEID=\"" << features_->GetUtteranceInfo().source_id
                <<           "\" SUBID=\"" << features_->GetUtteranceInfo().sub_id
                <<             "\" SEC=\"" << features_->GetUtteranceInfo().tv_sec
                <<            "\" USEC=\"" << features_->GetUtteranceInfo().tv_usec
                <<            "\" UUID=\"" << features_->GetID() << "\"";
      }
      //else if(features_->is_file_input() || features_->is_mfcnet_input()){
      else{
        ss_sock << "<RECOGOUT SOURCEID=\"" << features_->GetSourceInfo().source_id << "\"";
      }
      if(lm_name_ != ""){ ss_sock << " NAME=\"" << lm_name_ << "\""; }
      ss_sock << std::fixed << std::setprecision(6)
              << ">\n"
              << " <RHYPO RANK=\"" << 1
              <<     "\" SCORE=\"" << -(double)(weight.Value1() + weight.Value2())
              <<   "\" AMSCORE=\"" << -(double)weight.Value2()
              <<   "\" LMSCORE=\"" << -(double)weight.Value1() << "\">\n";

      size_t i = 0;
      while(i < words.size()){
        ss_sock << std::fixed << std::setprecision(3)
                << "  <WHYPO WORD=\"" << str_words[i]
                <<    "\" CLASSID=\"" << str_classids[i]
                <<      "\" PHONE=\"" << ""
                <<         "\" CM=\"" << 1.0f << "\"/>\n";
        i++;
      }
      ss_sock << " </RHYPO>\n</RECOGOUT>\n" << delimiter_sym_;

      result_client_->broadcast(ss_sock.str());
    }

    // Write result for stdout
    {
      if(enable_print_result_ && enable_print_lat_combine_){
        std::stringstream ss_cout;

        if(emulate_julius_){
          if(features_){
            if(features_->is_ex_mfcnet_input()){
              ss_cout << std::fixed << std::setprecision(6)
                      <<   "client_id = " << features_->GetUtteranceInfo().client_id
                      << ", mic_index = " << features_->GetUtteranceInfo().mic_index
                      << ", source_id = " << features_->GetUtteranceInfo().source_id
                      <<    ", sub_id = " << features_->GetUtteranceInfo().sub_id
                      <<   ", azimuth = " << features_->GetUtteranceInfo().azimuth
                      << ", elevation = " << features_->GetUtteranceInfo().elevation
                      <<       ", sec = " << features_->GetUtteranceInfo().tv_sec
                      <<      ", usec = " << features_->GetUtteranceInfo().tv_usec
                      <<      ", uuid = " << features_->GetID() << "\n";
            }
            //else if(features_->is_file_input() || features_->is_mfcnet_input()){
            else{
              ss_cout << std::fixed << std::setprecision(6)
                      <<   "source_id = " << features_->GetSourceInfo().source_id
                      <<   ", azimuth = " << features_->GetSourceInfo().azimuth
                      << ", elevation = " << features_->GetSourceInfo().elevation
                      <<       ", sec = " << features_->GetSourceInfo().tv_sec
                      <<      ", usec = " << features_->GetSourceInfo().tv_usec << "\n";
            }
          }
          else{
            ss_cout << "source_id = -1, azimuth = 0.000000, elevation = 0.000000, sec = 0, usec = 0 \n";
          }

          ss_cout << "### Recognition: Rescoring pass \n"
                  << "STAT: 00 \n"
                  << "sentence1: ";
          {
            size_t i = 0;
            while(i < words.size()){ ss_cout << str_words[i++] << " "; }
          }
          ss_cout << "\n";

          ss_cout << "wseq1: ";
          {
            size_t i = 0;
            while(i < words.size()){ ss_cout << str_classids[i++] << " "; }
          }
          ss_cout << "\n";

          ss_cout << "phseq1: ";
          if(phone_syms_ != NULL){
            size_t i = 0;
            while(i < phones_aligned.size()){
              kaldi::int32 tstate = trans_model_.TransitionIdToTransitionState(phones_aligned[i]);
              kaldi::int32 phone = trans_model_.TransitionStateToPhone(tstate);
              std::string s = phone_syms_->Find(phone);
              if(s == ""){ KALDI_ERR << "Phone-id " << phone << " not in symbol table."; }
              ss_cout << s << " ";
              i++;
            }
          }
          ss_cout << "\n";

          ss_cout << "cmscore1: ";
          {
            size_t i = 0;
            while(i < words.size()){
              ss_cout << std::fixed << std::setprecision(3) << 1.0f << " ";
              i++;
            }
          }
          ss_cout << "\n";

          ss_cout << std::fixed << std::setprecision(6)
                  << "score1: " << -(double)(weight.Value1() + weight.Value2())
                  << " ( AM: " << -(double)weight.Value2() << ", LM: " << -(double)weight.Value1() << " )";
        }
        else{
          ss_cout << "result: ";
          {
            size_t i = 0;
            while(i < words.size()){ ss_cout << str_classids[i++] << " "; }
          }
          ss_cout << "\n";

          ss_cout << "phone_seq: ";
          if(phone_syms_ != NULL){
            size_t i = 0;
            while(i < phones_aligned.size()){
              kaldi::int32 tstate = trans_model_.TransitionIdToTransitionState(phones_aligned[i]);
              kaldi::int32 phone = trans_model_.TransitionStateToPhone(tstate);
              std::string s = phone_syms_->Find(phone);
              if(s == ""){ KALDI_ERR << "Phone-id " << phone << " not in symbol table."; }
              ss_cout << s << " ";
              i++;
            }
          }
          ss_cout << "\n";

          ss_cout << std::fixed << std::setprecision(6)
                  <<     "score=" << -(double)(weight.Value1() + weight.Value2())
                  << ", amscore=" << -(double)weight.Value2()
                  << ", lmscore=" << -(double)weight.Value1();
        }

        out << ss_cout.str() << std::endl;
      }
    }
  }
}

} // namespace nnet3
} // namespace kaldi

