// pidstat.hpp: Display per-process statistics.

#ifndef MP_PIDSTAT_H_
#define MP_PIDSTAT_H_

#ifdef HAVE_LINUX_SCHED_H
#include <linux/sched.h>
#endif // HAVE_SYS_LINUX_SCHED_H

// Check definitions and settings defined in <linux/sched.h>
#ifndef SCHED_NORMAL
#define SCHED_NORMAL  0
#endif
#ifndef SCHED_FIFO
#define SCHED_FIFO  1
#endif
#ifndef SCHED_RR
#define SCHED_RR  2
#endif
#ifndef SCHED_BATCH
#define SCHED_BATCH  3
#endif
// SCHED_ISO is not yet implemented
#ifndef SCHED_IDLE
#define SCHED_IDLE  5
#endif
#ifndef SCHED_DEADLINE
#define SCHED_DEADLINE  6
#endif

#ifdef HAVE_SYS_PARAM_H
#include <sys/param.h>
#endif // HAVE_SYS_PARAM_H

// Check definitions and settings defined in <sys/param.h>
#ifndef _POSIX_ARG_MAX
#define _POSIX_ARG_MAX  128;
#endif
#ifndef LOGIN_NAME_MAX
#define LOGIN_NAME_MAX  32;
#endif

#include "stat_print_base.hpp"
#include <thread>

namespace multiplatform {

class resource_tracker {
// ActivitiesFlag
constexpr static unsigned int P_A_CPU   = 0x01;

constexpr bool DISPLAY_CPU(unsigned int l_flag){ return ((l_flag & P_A_CPU) == P_A_CPU); }

// TASK/CHILD
constexpr static unsigned int P_NULL   = 0x00;
constexpr static unsigned int P_TASK   = 0x01;
constexpr static unsigned int P_CHILD  = 0x02;

constexpr bool DISPLAY_TASK_STATS(unsigned int l_flag){ return ((l_flag & P_TASK) == P_TASK); }
constexpr bool DISPLAY_CHILD_STATS(unsigned int l_flag){ return ((l_flag & P_CHILD) == P_CHILD); }

constexpr static unsigned int P_D_PID        = 0x0001;
constexpr static unsigned int P_D_ALL_PID    = 0x0002;
constexpr static unsigned int P_D_ACTIVE_PID = 0x0010;
constexpr static unsigned int P_D_TID        = 0x0020;
constexpr static unsigned int P_D_UNIT       = 0x0800;

constexpr bool DISPLAY_PID(unsigned int l_flag){ return ((l_flag & P_D_PID) == P_D_PID); }
constexpr bool DISPLAY_ALL_PID(unsigned int l_flag){ return ((l_flag & P_D_ALL_PID) == P_D_ALL_PID); }
constexpr bool DISPLAY_ACTIVE_PID(unsigned int l_flag){ return ((l_flag & P_D_ACTIVE_PID) == P_D_ACTIVE_PID); }
constexpr bool DISPLAY_TID(unsigned int l_flag){ return ((l_flag & P_D_TID) == P_D_TID); }
constexpr bool DISPLAY_UNIT(unsigned int l_flag){ return ((l_flag & P_D_UNIT) == P_D_UNIT); }

// Per-process flags
constexpr static unsigned int F_NO_PID_IO     = 0x01;
constexpr static unsigned int F_NO_PID_FD     = 0x02;
constexpr static unsigned int F_PID_DISPLAYED = 0x04;

constexpr bool NO_PID_IO(unsigned int l_flag){ return ((l_flag & F_NO_PID_IO) == F_NO_PID_IO); }
constexpr bool NO_PID_FD(unsigned int l_flag){ return ((l_flag & F_NO_PID_FD) == F_NO_PID_FD); }
constexpr bool IS_PID_DISPLAYED(unsigned int l_flag){ return ((l_flag & F_PID_DISPLAYED) == F_PID_DISPLAYED); }

constexpr void PRINT_ID_HDR(char *l_timestamp, unsigned int l_flag){
  printf("\n%-11s", l_timestamp);
  printf("   UID");
  printf(DISPLAY_TID(l_flag) ? "      TGID       TID" : "       PID");
}



constexpr static unsigned int POLICY_NORMAL   = SCHED_NORMAL;
constexpr static unsigned int POLICY_FIFO     = SCHED_FIFO;
constexpr static unsigned int POLICY_RR       = SCHED_RR;
constexpr static unsigned int POLICY_BATCH    = SCHED_BATCH;
constexpr static unsigned int POLICY_IDLE     = SCHED_IDLE;
constexpr static unsigned int POLICY_DEADLINE = SCHED_DEADLINE;

constexpr const char * GET_POLICY(unsigned int p){
  return (p == POLICY_NORMAL   ? "NORMAL" :
         (p == POLICY_FIFO     ? "FIFO" :
         (p == POLICY_RR       ? "RR" :
         (p == POLICY_BATCH    ? "BATCH" :
         (p == POLICY_IDLE     ? "IDLE" :
         (p == POLICY_DEADLINE ? "DEADLN" :
         "?"))))));
}


constexpr static size_t MAX_COMM_LEN = _POSIX_ARG_MAX;
constexpr static size_t MAX_USER_LEN = LOGIN_NAME_MAX;


struct pid_stats {
  uint64_t read_bytes;
  uint64_t write_bytes;
  uint64_t cancelled_write_bytes;
  uint64_t blkio_swapin_delays;
  uint64_t minflt;
  uint64_t cminflt;
  uint64_t majflt;
  uint64_t cmajflt;
  uint64_t utime;
  int64_t  cutime;
  uint64_t stime;
  int64_t  cstime;
  uint64_t gtime;
  int64_t  cgtime;
  uint64_t wtime;
  uint64_t vsz;
  uint64_t rss;
  uint64_t nvcsw;
  uint64_t nivcsw;
  uint64_t stack_size;
  uint64_t stack_ref;
  unsigned int  processor;
  unsigned int  priority;
  unsigned int  policy;
  unsigned int  threads;
  unsigned int  fd_nr;
};

struct st_pid {
  uint64_t total_vsz;
  uint64_t total_rss;
  uint64_t total_stack_size;
  uint64_t total_stack_ref;
  uint64_t total_threads;
  uint64_t total_fd_nr;
  pid_t    pid;
  uid_t    uid;
  int      exist;  // TRUE if PID exists
  unsigned int flags;
  unsigned int rt_asum_count;
  unsigned int rc_asum_count;
  unsigned int uc_asum_count;
  unsigned int tf_asum_count;
  unsigned int sk_asum_count;
  unsigned int delay_asum_count;
  struct pid_stats *pstats[3];
  struct st_pid    *tgid;  // If current task is a TID, pointer to its TGID. NULL otherwise.
  struct st_pid    *next;
  char     comm[MAX_COMM_LEN];
  char     cmdline[MAX_COMM_LEN];
};


public:
  resource_tracker(): timer_th(nullptr), runner_th(nullptr) {
    timer_update = false; loop_continue = true; loop_pause = true; dprint = false;
  }
  ~resource_tracker() {
    count = 0;
    loop_continue = false;
    if(timer_th != nullptr){ timer_th->detach(); delete timer_th; timer_th = nullptr; }
    if(runner_th != nullptr){ runner_th->detach(); delete runner_th; runner_th = nullptr; }
  }

  void timer();
  void pause();
  void runner();

  void usage(char *progname);
  //void alarm_handler(int sig);
  //void int_handler(int sig);
  void sfree_pid(struct st_pid **plist, int force);
  void set_pid_nonexistent(struct st_pid *plist);
  struct st_pid *add_list_pid(struct st_pid **plist, pid_t pid, pid_t tgid);
  void read_proc_meminfo(void);
  int read_proc_pid_stat(pid_t pid, struct st_pid *plist,
           unsigned int *thread_nr, pid_t tgid, int curr);
  int read_proc_pid_sched(pid_t pid, struct st_pid *plist, pid_t tgid, int curr);
  int read_proc_pid_status(pid_t pid, struct st_pid *plist, pid_t tgid, int curr);
  int read_pid_stats(pid_t pid, struct st_pid *plist, unsigned int *thread_nr,
       pid_t tgid, int curr);
  void read_task_stats(pid_t pid, struct st_pid *plist, int curr);
  void read_stats(int curr);
  int get_pid_to_display(int prev, int curr, unsigned int activity, unsigned int pflag,
           struct st_pid *plist);
  int write_pid_task_cpu_stats(int prev, int curr, int disp_avg,
           char *prev_string, char *curr_string, uint64_t itv, uint64_t deltot_jiffies);
  int write_stats_core(int prev, int curr, int disp_avg,
                     char *prev_string, char *curr_string);
  void write_stats_avg(int curr);
  int write_stats(int curr);
  //void rw_pidstat_loop();

  //int standalone_test_main(int argc, char **argv);

  int stats_init(pid_t pid);
  //int stats_run();
  void update_interval(int64_t itv);
  void update_logging(bool enable);
  double get_cpu_usage();
  void stats_stop_request();
  int stats_free();

bool dprint;

int m_pid;

double m_user;
double m_system;
double m_guest;
double m_wait;
double m_total;

uint64_t hz;
unsigned int kb_shift;

uint64_t tot_jiffies[3] = {0, 0, 0};
uint64_t uptime_cs[3] = {0, 0, 0};
struct st_pid *pid_list = NULL;

struct tm ps_tstamp[3];
char commstr[MAX_COMM_LEN];
char userstr[MAX_USER_LEN];
char procstr[MAX_COMM_LEN];

int cpu_nr = 0;      // Nb of processors on the machine
uint64_t tlmkb;    // Total memory in kB
int64_t interval = -1;
int64_t count = 0;
unsigned int pidflag = 0;  // General flags
unsigned int tskflag = 0;  // TASK/CHILD stats
unsigned int actflag = 0;  // Activity flag

//struct sigaction alrm_act, int_act, chld_act;
int signal_caught = 0;


std::atomic<bool> loop_continue;
std::atomic<bool> loop_pause;
std::atomic<bool> timer_update;
  std::thread *timer_th;
  std::thread *runner_th;

}; // class resource_tracker

} // namespace multiplatform


class resource_monitor {
 public:
  resource_monitor(pid_t pid);
  virtual ~resource_monitor();

  virtual void set_mode(bool l_enable_cut_off, bool l_enable_resource_logging,
                        int l_check_interval,
                        int l_cpu_usage_threshold, int l_memory_usage_threshold);
  //virtual void set_mode(bool l_enable_cut_off, bool l_enable_resource_logging,
  //                      int l_check_interval,
  //                      int l_cpu_usage_threshold, int l_memory_usage_threshold) {
  //  enable_feature = l_enable_cut_off;
  //  enable_logging = l_enable_resource_logging;
  //  interval = (l_check_interval < 0) ? 0 : l_check_interval;
  //  cpu_thresh = l_cpu_usage_threshold;
  //  mem_thresh = l_memory_usage_threshold;
  //}

  //void run();
  //void run(){
  //}

  virtual bool is_cpu_warning();
  //virtual bool is_cpu_warning() { return false; }
  //virtual bool is_mem_warning() { return false; }

  multiplatform::resource_tracker *rt;
  pid_t m_pid; // for backup
  //std::thread *th;
  bool enable_feature;
  //bool enable_logging;
  int interval;
  int cpu_thresh;
  int mem_thresh;
};

#endif // MP_PIDSTAT_H_

