// pidstat.cpp: pidstat like test command by C++11 coding, Report statistics for Linux tasks

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <unistd.h>
#include <fcntl.h>
#include <signal.h>
#include <dirent.h>
#include <ctype.h>
#include <sys/types.h>
#include <pwd.h>
#include <sys/utsname.h>
#include <regex.h>
#include <inttypes.h>

#include <sstream>

#include "pidstat.hpp"
#include "read_stats.hpp"

namespace multiplatform {

uint64_t hz;
unsigned int kb_shift;

uint64_t tot_jiffies[3] = {0, 0, 0};
uint64_t uptime_cs[3] = {0, 0, 0};
struct st_pid *pid_list = NULL;

struct tm ps_tstamp[3];
char commstr[MAX_COMM_LEN];
char userstr[MAX_USER_LEN];
char procstr[MAX_COMM_LEN];

int cpu_nr = 0;      // Nb of processors on the machine
uint64_t tlmkb;    // Total memory in kB
int64_t interval = -1;
int64_t count = 0;
unsigned int pidflag = 0;  // General flags
unsigned int tskflag = 0;  // TASK/CHILD stats
unsigned int actflag = 0;  // Activity flag

struct sigaction alrm_act, int_act, chld_act;
int signal_caught = 0;

int dplaces_nr = -1;    // Number of decimal places

void usage(char *progname)
{
  std::stringstream ss;
  ss << "Usage:\n  "
     << progname << " [ options ] [ <interval> [ <count> ] ] [ -e <program> <args> ]\n"
        "Options are:\n"
        "  [ -d ] [ -H ] [ -h ] [ -I ] [ -l ] [ -R ] [ -r ] [ -s ] [ -t ]\n"
        "  [ -U [ <username> ] ] [ -u ] [ -V ] [ -v ] [ -w ] [ -C <command> ]\n"
        "  [ -G <process_name> ] [ -p { <pid> [,...] | SELF | ALL } ]\n"
        "  [ -T { TASK | CHILD | ALL } ] [ --dec={ 0 | 1 | 2 } ] [ --human ]\n";
  fprintf(stderr, "%s", ss.str().c_str());
  exit(1);
}


void alarm_handler(int sig)
{
  alarm(interval);
}


void int_handler(int sig)
{
  signal_caught = 1;
}


void sfree_pid(struct st_pid **plist, int force)
{
  int i;
  struct st_pid *p;

  while(*plist != NULL){
    p = *plist;
    if(!p->exist || force){
      *plist = p->next;
      for(i = 0; i < 3; i++){
        if(p->pstats[i]){ free(p->pstats[i]); }
      }
      free(p);
    }
    else{ plist = &(p->next); }
  }
}


void set_pid_nonexistent(struct st_pid *plist)
{
  while(plist != NULL){
    plist->exist = FALSE;
    plist = plist->next;
  }
}


struct st_pid *add_list_pid(struct st_pid **plist, pid_t pid, pid_t tgid)
{
  struct st_pid *p, *ps, *tgid_p = NULL;
  int i;
  int tgid_found = FALSE;

  if(!pid){ return NULL; }

  if(!tgid){
    // Add a true PID to the list.
    // Add it in ascending order, not taking into account other TIDs.
    while(*plist != NULL){
      p = *plist;
      // PID found in list
      if(!p->tgid && (p->pid == pid)){ return p; }

      // Stop now to insert PID in list
      if(!p->tgid && (p->pid > pid)){ break; }

      plist = &(p->next);
    }
  }
  else{
    // PID is a TID.
    // It will be inserted in ascending order immediately following its TGID.
    while(*plist != NULL){
      p = *plist;
      if(p->pid == tgid){
        // TGID found in list
        tgid_found = TRUE;
        tgid_p = p;
        break;
      }

      plist = &(p->next);
    }
    // TGID not found: Stop now
    if(!tgid_found){ return NULL; }

    plist = &(p->next);
    while(*plist != NULL){
      p = *plist;
      // TID found in list
      if((p->tgid == tgid_p) && (p->pid == pid)){ return p; }

      // Stop now to insert TID in list
      if((p->tgid == tgid_p) && (p->pid > pid)){ break; }

      // End of TID list: insert TID here
      if(p->tgid != tgid_p){ break; }

      plist = &(p->next);
    }
  }

  // PID not found
  ps = *plist;

  // Add PID to the list
  if((*plist = (struct st_pid *) malloc(sizeof(struct st_pid))) == NULL){
    perror("malloc");
    exit(4);
  }
  memset(*plist, 0, sizeof(struct st_pid));

  p = *plist;
  for(i = 0; i < 3; i++){
    if((p->pstats[i] = (struct pid_stats *) malloc(sizeof(struct pid_stats))) == NULL){
      perror("malloc");
      exit(4);
    }
    memset(p->pstats[i], 0, sizeof(struct pid_stats));
  }
  p->pid = pid;
  p->next = ps;
  if(tgid_p){ p->tgid = tgid_p; }

  return p;
}

/*
void pid_lists::sfree_pid(int force)
{
  int i;
  struct st_pid *p;

  while(*plist != NULL){
    p = *plist;
    if(!p->exist || force){
      *plist = p->next;
      for(i = 0; i < 3; i++){
        if(p->pstats[i]){ free(p->pstats[i]); }
      }
      free(p);
    }
    else{ plist = &(p->next); }
  }
}


void pid_lists::set_pid_nonexistent(struct st_pid *plist)
{
  while(plist != NULL){
    plist->exist = FALSE;
    plist = plist->next;
  }
}


struct st_pid *add_list_pid(struct st_pid **plist, pid_t pid, pid_t tgid)
{
  struct st_pid *p, *ps, *tgid_p = NULL;
  int i;
  int tgid_found = FALSE;

  if(!pid){ return NULL; }

  if(!tgid){
    // Add a true PID to the list.
    // Add it in ascending order, not taking into account other TIDs.
    while(*plist != NULL){
      p = *plist;
      // PID found in list
      if(!p->tgid && (p->pid == pid)){ return p; }

      // Stop now to insert PID in list
      if(!p->tgid && (p->pid > pid)){ break; }

      plist = &(p->next);
    }
  }
  else{
    // PID is a TID.
    // It will be inserted in ascending order immediately following its TGID.
    while(*plist != NULL){
      p = *plist;
      if(p->pid == tgid){
        // TGID found in list
        tgid_found = TRUE;
        tgid_p = p;
        break;
      }

      plist = &(p->next);
    }
    // TGID not found: Stop now
    if(!tgid_found){ return NULL; }

    plist = &(p->next);
    while(*plist != NULL){
      p = *plist;
      // TID found in list
      if((p->tgid == tgid_p) && (p->pid == pid)){ return p; }

      // Stop now to insert TID in list
      if((p->tgid == tgid_p) && (p->pid > pid)){ break; }

      // End of TID list: insert TID here
      if(p->tgid != tgid_p){ break; }

      plist = &(p->next);
    }
  }

  // PID not found
  ps = *plist;

  // Add PID to the list
  if((*plist = (struct st_pid *) malloc(sizeof(struct st_pid))) == NULL){
    perror("malloc");
    exit(4);
  }
  memset(*plist, 0, sizeof(struct st_pid));

  p = *plist;
  for(i = 0; i < 3; i++){
    if((p->pstats[i] = (struct pid_stats *) malloc(sizeof(struct pid_stats))) == NULL){
      perror("malloc");
      exit(4);
    }
    memset(p->pstats[i], 0, sizeof(struct pid_stats));
  }
  p->pid = pid;
  p->next = ps;
  if(tgid_p){ p->tgid = tgid_p; }

  return p;
}
*/

void check_flags(void)
{
  unsigned int act = 0;

  // Display CPU usage for active tasks by default
  if(!actflag){ actflag |= P_A_CPU; }

  // If no PIDs nor -p ALL entered on the command line then
  // only active PIDs will be displayed.
  if(!DISPLAY_PID(pidflag)){ pidflag |= P_D_ACTIVE_PID + P_D_ALL_PID; }

  if(!tskflag){ tskflag |= P_TASK; }

  // Check that requested activities are available
  if(DISPLAY_TASK_STATS(tskflag)){
    act |= P_A_CPU + P_A_MEM + P_A_IO + P_A_CTXSW + P_A_STACK + P_A_KTAB + P_A_RT;
  }
  if(DISPLAY_CHILD_STATS(tskflag)){ act |= P_A_CPU + P_A_MEM; }

  actflag &= act;

  if(!actflag){
    fprintf(stderr, "Requested activities not available\n");
    exit(1);
  }
}


char *get_tcmd(struct st_pid *plist)
{
  // Option "-l" used
  if(DISPLAY_CMDLINE(pidflag) && strlen(plist->cmdline) && !plist->tgid){
    return plist->cmdline;
  }
  else{
    return plist->comm;
  }
}


void print_comm(struct st_pid *plist)
{
  multiplatform::cprint_support *cs = new multiplatform::cprint_support();
  char *p;

  // Get pointer on task's command string
  p = get_tcmd(plist);

  if(plist->tgid){
    if(IS_PID_DISPLAYED(plist->tgid->flags)){
      cs->cprintf_s(cprint_support::IS_ZERO, (char *) "  |__%s\n", p);
    }
    else{
      // Its TGID has not been displayed
      cs->cprintf_s(cprint_support::IS_STR, (char *) "  (%s)", plist->tgid->comm);
      cs->cprintf_s(cprint_support::IS_ZERO, (char *) "__%s\n", p);

      // We can now consider this has been the case
      plist->tgid->flags |= F_PID_DISPLAYED;
    }
  }
  else{
    cs->cprintf_s(cprint_support::IS_STR, (char *) "  %s\n", p);
  }
  delete cs;
}


void read_proc_meminfo(void)
{
  struct stats_memory st_mem;

  memset(&st_mem, 0, sizeof(struct stats_memory));
  {
    multiplatform::read_statistics *res = new multiplatform::read_statistics();
    res->read_meminfo(&st_mem);
    delete res;
  }
  tlmkb = st_mem.tlmkb;
}


int read_proc_pid_stat(pid_t pid, struct st_pid *plist,
           unsigned int *thread_nr, pid_t tgid, int curr)
{
  constexpr static char PID_STAT[] = "/proc/%u/stat";
  constexpr static char TASK_STAT[] = "/proc/%u/task/%u/stat";
  int fd, sz, rc;
  size_t commsz;
  char filename[128];
  static char buffer[1024 + 1];
  char *start, *end;
  struct pid_stats *pst = plist->pstats[curr];

  if(tgid){ sprintf(filename, TASK_STAT, tgid, pid); }
  else{ sprintf(filename, PID_STAT, pid); }

  // No such process
  if((fd = open(filename, O_RDONLY)) < 0){ return 1; }

  sz = read(fd, buffer, 1024);
  close(fd);
  if(sz <= 0){ return 1; }
  buffer[sz] = '\0';

  if((start = strchr(buffer, '(')) == NULL){ return 1; }
  start += 1;
  if((end = strrchr(start, ')')) == NULL){ return 1; }
  commsz = end - start;
  if(commsz >= MAX_COMM_LEN){ return 1; }
  memcpy(plist->comm, start, commsz);
  plist->comm[commsz] = '\0';
  start = end + 2;

  rc = sscanf(start,
        "%*s %*d %*d %*d %*d %*d %*u"
        " %" PRIu64 " %" PRIu64 " %" PRIu64 " %" PRIu64
        " %" PRIu64 " %" PRIu64 " %" PRId64 " %" PRId64
        " %*d %*d %u %*u %*d %" PRIu64 " %" PRIu64
        " %*u %*u %*u %*u %*u %*u %*u %*u %*u %*u %*u %*u %*u %*u"
        " %u %u %u %" PRIu64 " %" PRIu64 " %" PRId64 "\n",
        &pst->minflt, &pst->cminflt, &pst->majflt, &pst->cmajflt,
        &pst->utime,  &pst->stime, &pst->cutime, &pst->cstime,
        thread_nr, &pst->vsz, &pst->rss,
        &pst->processor, &pst->priority, &pst->policy,
        &pst->blkio_swapin_delays, &pst->gtime, &pst->cgtime);

  if(rc < 15){ return 1; }

  // gtime and cgtime fields are unavailable in file
  if(rc < 17){ pst->gtime = pst->cgtime = 0; }

  // Convert to kB
  {
    multiplatform::read_statistics *res = new multiplatform::read_statistics();
    res->get_kb_shift();
    pst->vsz >>= 10;
    pst->rss = res->PG_TO_KB(pst->rss);
    delete res;
  }

  return 0;
}


int read_proc_pid_sched(pid_t pid, struct st_pid *plist, pid_t tgid, int curr)
{
  constexpr static char PID_SCHED[] = "/proc/%u/schedstat";
  constexpr static char TASK_SCHED[] = "/proc/%u/task/%u/schedstat";
  int fd, sz, rc = 0;
  char filename[128];
  static char buffer[1024 + 1];
  uint64_t wtime = 0;
  struct pid_stats *pst = plist->pstats[curr];

  if(tgid){ sprintf(filename, TASK_SCHED, tgid, pid); }
  else{ sprintf(filename, PID_SCHED, pid); }

  if((fd = open(filename, O_RDONLY)) >= 0){
    // schedstat file found for process
    sz = read(fd, buffer, 1024);
    close(fd);
    if(sz > 0){ buffer[sz] = '\0'; rc = sscanf(buffer, "%*u %" PRIu64 " %*d\n", &wtime); }
  }

  // Convert ns to jiffies
  pst->wtime = wtime * hz / 1000000000;

  if(rc < 1){ return 1; }

  return 0;
}


int read_proc_pid_status(pid_t pid, struct st_pid *plist, pid_t tgid, int curr)
{
  constexpr static char PID_STATUS[] = "/proc/%u/status";
  constexpr static char TASK_STATUS[] = "/proc/%u/task/%u/status";
  FILE *fp;
  char filename[128], line[256];
  struct pid_stats *pst = plist->pstats[curr];

  if(tgid){ sprintf(filename, TASK_STATUS, tgid, pid); }
  else{ sprintf(filename, PID_STATUS, pid); }

  // No such process
  if((fp = fopen(filename, "r")) == NULL){ return 1; }

  while(fgets(line, sizeof(line), fp) != NULL){
    if(!strncmp(line, "Uid:", 4)){
      sscanf(line + 5, "%u", &plist->uid);
    }
    else if(!strncmp(line, "Threads:", 8)){
      sscanf(line + 9, "%u", &pst->threads);
    }
    else if(!strncmp(line, "voluntary_ctxt_switches:", 24)){
      sscanf(line + 25, "%" PRIu64, &pst->nvcsw);
    }
    else if(!strncmp(line, "nonvoluntary_ctxt_switches:", 27)){
      sscanf(line + 28, "%" PRIu64, &pst->nivcsw);
    }
  }

  fclose(fp);

  return 0;
}


int read_proc_pid_smap(pid_t pid, struct st_pid *plist, pid_t tgid, int curr)
{
  constexpr static char PID_SMAP[] = "/proc/%u/smaps";
  constexpr static char TASK_SMAP[] = "/proc/%u/task/%u/smaps";
  FILE *fp;
  char filename[128], line[256];
  int state = 0;
  struct pid_stats *pst = plist->pstats[curr];

  if(tgid){ sprintf(filename, TASK_SMAP, tgid, pid); }
  else{ sprintf(filename, PID_SMAP, pid); }

  // No such process
  if((fp = fopen(filename, "rt")) == NULL){ return 1; }

  while((state < 3) && (fgets(line, sizeof(line), fp) != NULL)){
    switch(state){
      case 0:
        if(strstr(line, "[stack]")){
          state = 1;
        }
        break;
      case 1:
        if(strstr(line, "Size:")){
          sscanf(line + sizeof("Size:"), "%" PRIu64, &pst->stack_size);
          state = 2;
        }
        break;
      case 2:
        if(strstr(line, "Referenced:")){
          sscanf(line + sizeof("Referenced:"), "%" PRIu64, &pst->stack_ref);
          state = 3;
        }
        break;
    }
  }

  fclose(fp);

  return 0;
}


int read_proc_pid_cmdline(pid_t pid, struct st_pid *plist, pid_t tgid)
{
  constexpr static char PID_CMDLINE[] = "/proc/%u/cmdline";
  constexpr static char TASK_CMDLINE[] = "/proc/%u/task/%u/cmdline";
  FILE *fp;
  char filename[128], line[MAX_COMM_LEN];
  size_t len;
  int i, found = FALSE;

  if(tgid){ sprintf(filename, TASK_CMDLINE, tgid, pid); }
  else{ sprintf(filename, PID_CMDLINE, pid); }

  // No such process
  if((fp = fopen(filename, "r")) == NULL){ return 1; }

  memset(line, 0, MAX_COMM_LEN);

  len = fread(line, 1, sizeof(line) - 1, fp);
  fclose(fp);

  if(len){
    for(i = len - 2; i >= 0; i--){
      if(line[i]){ found = TRUE; }
      else if(found){ line[i] = ' '; }
    }
    strncpy(plist->cmdline, line, sizeof(plist->cmdline) - 1);
    plist->cmdline[sizeof(plist->cmdline) - 1] = '\0';
  }
  else{
    // proc/.../cmdline was empty
    plist->cmdline[0] = '\0';
  }
  return 0;
}


int read_proc_pid_io(pid_t pid, struct st_pid *plist, pid_t tgid, int curr)
{
  constexpr static char PID_IO[] = "/proc/%u/io";
  constexpr static char TASK_IO[] = "/proc/%u/task/%u/io";
  FILE *fp;
  char filename[128], line[256];
  struct pid_stats *pst = plist->pstats[curr];

  if(tgid){ sprintf(filename, TASK_IO, tgid, pid); }
  else{ sprintf(filename, PID_IO, pid); }

  if((fp = fopen(filename, "r")) == NULL){
    // No such process... or file non existent!
    plist->flags |= F_NO_PID_IO;
    // Also returns 0 since io stats file doesn't necessarily exist,
    // depending on the kernel version used.
    return 0;
  }

  while(fgets(line, sizeof(line), fp) != NULL){
    if(!strncmp(line, "read_bytes:", 11)){
      sscanf(line + 12, "%" PRIu64, &pst->read_bytes);
    }
    else if(!strncmp(line, "write_bytes:", 12)){
      sscanf(line + 13, "%" PRIu64, &pst->write_bytes);
    }
    else if(!strncmp(line, "cancelled_write_bytes:", 22)){
      sscanf(line + 23, "%" PRIu64, &pst->cancelled_write_bytes);
    }
  }

  fclose(fp);

  plist->flags &= ~F_NO_PID_IO;

  return 0;
}


int read_proc_pid_fd(pid_t pid, struct st_pid *plist, pid_t tgid, int curr)
{
  constexpr static char PID_FD[] = "/proc/%u/fd";
  constexpr static char TASK_FD[] = "/proc/%u/task/%u/fd";
  DIR *dir;
  struct dirent *drp;
  char filename[128];
  struct pid_stats *pst = plist->pstats[curr];

  if(tgid){ sprintf(filename, TASK_FD, tgid, pid); }
  else{ sprintf(filename, PID_FD, pid); }

  if((dir = opendir(filename)) == NULL){
    // Cannot read fd directory
    plist->flags |= F_NO_PID_FD;
    return 0;
  }

  pst->fd_nr = 0;

  // Count number of entries if fd directory
  while((drp = readdir(dir)) != NULL){
    if(isdigit(drp->d_name[0])){ (pst->fd_nr)++; }
  }

  closedir(dir);

  plist->flags &= ~F_NO_PID_FD;

  return 0;
}


int read_pid_stats(pid_t pid, struct st_pid *plist, unsigned int *thread_nr,
       pid_t tgid, int curr)
{
  if(read_proc_pid_stat(pid, plist, thread_nr, tgid, curr)){ return 1; }

  // No need to test the return code here: Not finding
  // the schedstat files shouldn't make pidstat stop.
  read_proc_pid_sched(pid, plist, tgid, curr);

  if(DISPLAY_CMDLINE(pidflag) && !plist->cmdline[0]){
    if(read_proc_pid_cmdline(pid, plist, tgid)){ return 1; }
  }

  if(read_proc_pid_status(pid, plist, tgid, curr)){ return 1; }

  if(DISPLAY_STACK(actflag)){
    if(read_proc_pid_smap(pid, plist, tgid, curr)){ return 1; }
  }

  if(DISPLAY_KTAB(actflag)){
    if(read_proc_pid_fd(pid, plist, tgid, curr)){ return 1; }
  }

  // Assume that /proc/#/task/#/io exists!
  if(DISPLAY_IO(actflag)){ return (read_proc_pid_io(pid, plist, tgid, curr)); }

  return 0;
}


void read_task_stats(pid_t pid, struct st_pid *plist, int curr)
{
  constexpr static char PROC_TASK[] = "/proc/%u/task";
  DIR *dir;
  pid_t tid;
  struct dirent *drp;
  char filename[128];
  unsigned int thr_nr;
  struct st_pid *tlist;

  // Open /proc/#/task directory
  sprintf(filename, PROC_TASK, pid);
  if((dir = opendir(filename)) == NULL){ return; }

  while((drp = readdir(dir)) != NULL){
    if(!isdigit(drp->d_name[0])){ continue; }

    tid = atoi(drp->d_name);

    tlist = add_list_pid(&pid_list, tid, pid);
    if(!tlist){ continue; }
    tlist->exist = TRUE;

    if(read_pid_stats(tid, tlist, &thr_nr, pid, curr)){
      // Thread doesn't exist
      tlist->exist = FALSE;
    }
  }

  closedir(dir);
}


void read_stats(int curr)
{
  constexpr static char PROC[] = "/proc";
  DIR *dir;
  struct dirent *drp;
  unsigned int thr_nr;
  pid_t pid;
  struct st_pid *plist;
  struct stats_cpu *st_cpu;

  // Allocate two structures for CPU statistics.
  // No need to init them (done by read_stat_cpu() function).
  if((st_cpu = (struct stats_cpu *) malloc(sizeof(struct stats_cpu) * 2)) == NULL){
    perror("malloc");
    exit(4);
  }
  // Read statistics for CPUs "all"
  {
    multiplatform::read_statistics *res = new multiplatform::read_statistics();
    res->read_stat_cpu(st_cpu, 1);
    delete res;
  }

  // Compute the total number of jiffies spent by all processors.
  // NB: Don't add cpu_guest/cpu_guest_nice because cpu_user/cpu_nice already include them.
  tot_jiffies[curr] = st_cpu->cpu_user + st_cpu->cpu_nice +
          st_cpu->cpu_sys + st_cpu->cpu_idle +
          st_cpu->cpu_iowait + st_cpu->cpu_hardirq +
          st_cpu->cpu_steal + st_cpu->cpu_softirq;
  free(st_cpu);

  if(DISPLAY_ALL_PID(pidflag)){
    // Open /proc directory
    if((dir = opendir(PROC)) == NULL){
      perror("opendir");
      exit(4);
    }

    // Get directory entries
    while((drp = readdir(dir)) != NULL){
      if(!isdigit(drp->d_name[0])){ continue; }

      pid = atoi(drp->d_name);

      plist = add_list_pid(&pid_list, pid, 0);
      if(!plist){ continue; }
      plist->exist = TRUE;

      if(read_pid_stats(pid, plist, &thr_nr, 0, curr)){
        // PID has terminated
        plist->exist = FALSE;

      }
      else if(DISPLAY_TID(pidflag)){
        // Read stats for threads in task subdirectory
        read_task_stats(pid, plist, curr);
      }
    }

    closedir(dir);
  }

  else{
    // Read stats for each PID in the list
    for(plist = pid_list; plist != NULL; plist = plist->next){
      // Ignore TIDs.
      // The user can only enter PIDs on the command line.
      // If there is a TID then this is because the user has used option -t,
      // and the TID has been inserted in the list by read_task_stats() function below.
      if(plist->tgid){ continue; }

      if(read_pid_stats(plist->pid, plist, &thr_nr, 0, curr)){
        // PID has terminated
        plist->exist = FALSE;
      }
      else{
        plist->exist = TRUE;

        if(DISPLAY_TID(pidflag)){
          read_task_stats(plist->pid, plist, curr);
        }
      }
    }
  }

  // Free unused PID structures
  sfree_pid(&pid_list, FALSE);
}


int get_pid_to_display(int prev, int curr, unsigned int activity, unsigned int pflag,
           struct st_pid *plist)
{
  int rc;
  char *pc;
  regex_t regex;
  struct passwd *pwdent;
  struct pid_stats *pstc = plist->pstats[curr], *pstp = plist->pstats[prev];

  // PID no longer exists
  if(!plist->exist){ return 0; }

  if(!plist->tgid){
    // This is group leader: Set it as not displayed by default
    plist->flags &= ~F_PID_DISPLAYED;
  }

  if((DISPLAY_ALL_PID(pidflag) || DISPLAY_TID(pidflag)) && DISPLAY_ACTIVE_PID(pidflag)){
    int isActive = FALSE;

    // Check that it's an "active" process
    if(DISPLAY_CPU(activity)){
      // User time already includes guest time
      if((pstc->utime != pstp->utime) || (pstc->stime != pstp->stime)){
        isActive = TRUE;
      }
      else{
        // Process is not active but if we are showing child stats then we need to look there.
        if(DISPLAY_CHILD_STATS(pflag)){
          // User time already includes guest time
          if((pstc->cutime != pstp->cutime) || (pstc->cstime != pstp->cstime)){
            isActive = TRUE;
          }
        }
      }
    }

    if(DISPLAY_MEM(activity) && !isActive){
      if((pstc->minflt != pstp->minflt) || (pstc->majflt != pstp->majflt)){
        isActive = TRUE;
      }
      else{
        if(DISPLAY_TASK_STATS(pflag)){
          if((pstc->vsz != pstp->vsz) || (pstc->rss != pstp->rss)){
            isActive = TRUE;
          }
        }
        else if(DISPLAY_CHILD_STATS(pflag)){
          if((pstc->cminflt != pstp->cminflt) || (pstc->cmajflt != pstp->cmajflt)){
            isActive = TRUE;
          }
        }
      }
    }

    if(DISPLAY_STACK(activity) && !isActive){
      if((pstc->stack_size != pstp->stack_size) || (pstc->stack_ref != pstp->stack_ref)){
        isActive = TRUE;
      }
    }

    if(DISPLAY_IO(activity) && !isActive){
      if(pstc->blkio_swapin_delays != pstp->blkio_swapin_delays){
        isActive = TRUE;
      }
      // /proc/#/io file should exist to check I/O stats
      if(!(NO_PID_IO(plist->flags)) && !isActive){
        if((pstc->read_bytes  != pstp->read_bytes) ||
           (pstc->write_bytes != pstp->write_bytes) ||
           (pstc->cancelled_write_bytes != pstp->cancelled_write_bytes)){
          isActive = TRUE;
        }
      }
    }

    if(DISPLAY_CTXSW(activity) && !isActive){
      if((pstc->nvcsw  != pstp->nvcsw) || (pstc->nivcsw != pstp->nivcsw)){
        isActive = TRUE;
      }
    }

    if(DISPLAY_RT(activity) && !isActive){
      if((pstc->priority != pstp->priority) || (pstc->policy != pstp->policy)){
        isActive = TRUE;
      }
    }

    //                                        /proc/#/fd directory should be readable
    if(DISPLAY_KTAB(activity) && !isActive && !(NO_PID_FD(plist->flags))){
      if((pstc->threads != pstp->threads) || (pstc->fd_nr != pstp->fd_nr)){ isActive = TRUE; }
    }

    // If PID isn't active for any of the activities then return
    if(!isActive){ return -1; }
  }

  if(COMMAND_STRING(pidflag)){
    // Error in preparing regex structure
    if(regcomp(&regex, commstr, REG_EXTENDED | REG_NOSUB) != 0){ return -1; }

    pc = get_tcmd(plist);  // Get pointer on task's command string
    rc = regexec(&regex, pc, 0, NULL, 0);
    regfree(&regex);

    // regex pattern not found in command name
    if(rc){ return -1; }
  }

  if(PROCESS_STRING(pidflag)){
    if(!plist->tgid){
      // This PID is a process ("thread group leader")
      // Error in preparing regex structure
      if(regcomp(&regex, procstr, REG_EXTENDED | REG_NOSUB) != 0){ return -1; }

      pc = get_tcmd(plist);  // Get pointer on task's command string
      rc = regexec(&regex, pc, 0, NULL, 0);
      regfree(&regex);

      // regex pattern not found in command name
      if(rc){ return -1; }
    }
    // This pid is a thread and is not part of a process to display
    else if(!IS_PID_DISPLAYED(plist->tgid->flags)){ return -1; }
  }

  if(USER_STRING(pidflag)){
    if((pwdent = getpwuid(plist->uid)) != NULL){
      // This PID doesn't belong to user
      if(strcmp(pwdent->pw_name, userstr)){ return -1; }
    }
  }

  plist->flags |= F_PID_DISPLAYED;
  return 1;
}


void print_line_id(char *timestamp, struct st_pid *plist)
{
  multiplatform::cprint_support *cs = new multiplatform::cprint_support();
  char format[32];
  struct passwd *pwdent;

  printf("%-11s", timestamp);
  //__print_line_id(plist, '-');

  if(DISPLAY_USERNAME(pidflag) && ((pwdent = getpwuid(plist->uid)) != NULL)){
    cs->cprintf_in(cprint_support::IS_STR, (char *) " %8s", pwdent->pw_name, 0);
  }
  else{
    cs->cprintf_in(cprint_support::IS_INT, (char *) " %5d", (char *) "", plist->uid);
  }

  if(DISPLAY_TID(pidflag)){
    if(plist->tgid){
      // This is a TID
      if(IS_PID_DISPLAYED(plist->tgid->flags)){
        sprintf(format, "         %c %%9u", '-');
      }
      else{
        strcpy(format, " %9u");
        cs->cprintf_in(cprint_support::IS_INT, format, (char *) "", plist->tgid->pid);
      }
    }
    else{
      // This is a PID (TGID)
      sprintf(format, " %%9u         %c", '-');
    }
  }
  else{ strcpy(format, " %9u"); }

  cs->cprintf_in(cprint_support::IS_INT, format, (char *) "", plist->pid);
  delete cs;
}


int write_pid_task_all_stats(int prev, int curr, int dis,
        char *prev_string, char *curr_string, uint64_t itv, uint64_t deltot_jiffies)
{
  multiplatform::cprint_support *cs = new multiplatform::cprint_support();
  struct pid_stats *pstc, *pstp;
  struct st_pid *plist;
  char dstr[32];
  int again = 0;

  if(dis){
    PRINT_ID_HDR(prev_string, pidflag);
    if(DISPLAY_CPU(actflag)){ printf("    %%usr %%system  %%guest   %%wait    %%CPU   CPU"); }
    if(DISPLAY_MEM(actflag)){ printf("  minflt/s  majflt/s     VSZ     RSS   %%MEM"); }
    if(DISPLAY_STACK(actflag)){ printf(" StkSize  StkRef"); }
    if(DISPLAY_IO(actflag)){ printf("   kB_rd/s   kB_wr/s kB_ccwr/s iodelay"); }
    if(DISPLAY_CTXSW(actflag)){ printf("   cswch/s nvcswch/s"); }
    if(DISPLAY_KTAB(actflag)){ printf(" threads   fd-nr"); }
    if(DISPLAY_RT(actflag)){ printf(" prio policy"); }
    printf("  Command\n");
  }

  for(plist = pid_list; plist != NULL; plist = plist->next){
    if(get_pid_to_display(prev, curr, actflag, P_TASK, plist) <= 0){ continue; }

    print_line_id(curr_string, plist);

    pstc = plist->pstats[curr];
    pstp = plist->pstats[prev];

    if(DISPLAY_CPU(actflag)){
      cs->cprintf_pc(DISPLAY_UNIT(pidflag), 5, 7, 2,
          (pstc->utime - pstc->gtime) < (pstp->utime - pstp->gtime) ?
          0.0 :
          sp_value(pstp->utime - pstp->gtime, pstc->utime - pstc->gtime, itv * hz / 100),
          sp_value(pstp->stime,  pstc->stime, itv * hz / 100),
          sp_value(pstp->gtime,  pstc->gtime, itv * hz / 100),
          sp_value(pstp->wtime,  pstc->wtime, itv * hz / 100),
          // User time already includes guest time
          IRIX_MODE_OFF(pidflag) ?
          sp_value(pstp->utime + pstp->stime, pstc->utime + pstc->stime, deltot_jiffies) :
          sp_value(pstp->utime + pstp->stime, pstc->utime + pstc->stime, itv * hz / 100));

      cs->cprintf_in(cprint_support::IS_INT, (char *) "   %3d", (char *) "", pstc->processor);
    }

    if(DISPLAY_MEM(actflag)){
      cs->cprintf_f(cprint_support::NO_UNIT, 2, 9, 2,
          s_value(pstp->minflt, pstc->minflt, itv), s_value(pstp->majflt, pstc->majflt, itv));
      cs->cprintf_u64(DISPLAY_UNIT(pidflag) ? cprint_support::UNIT_KILOBYTE : cprint_support::NO_UNIT, 2, 7,
            (uint64_t) pstc->vsz, (uint64_t) pstc->rss);
      cs->cprintf_pc(DISPLAY_UNIT(pidflag), 1, 6, 2, tlmkb ? sp_value(0, pstc->rss, tlmkb) : 0.0);
    }

    if(DISPLAY_STACK(actflag)){
      cs->cprintf_u64(DISPLAY_UNIT(pidflag) ? cprint_support::UNIT_KILOBYTE : cprint_support::NO_UNIT, 2, 7,
            (uint64_t) pstc->stack_size, (uint64_t) pstc->stack_ref);
    }

    if(DISPLAY_IO(actflag)){
      if(!NO_PID_IO(plist->flags)){
        double rbytes, wbytes, cbytes;

        rbytes = s_value(pstp->read_bytes,  pstc->read_bytes, itv);
        wbytes = s_value(pstp->write_bytes, pstc->write_bytes, itv);
        cbytes = s_value(pstp->cancelled_write_bytes, pstc->cancelled_write_bytes, itv);
        if(!DISPLAY_UNIT(pidflag)){ rbytes /= 1024; wbytes /= 1024; cbytes /= 1024; }
        cs->cprintf_f(DISPLAY_UNIT(pidflag) ? cprint_support::UNIT_BYTE : cprint_support::NO_UNIT, 3, 9, 2, rbytes, wbytes, cbytes);
      }
      else{
        // Keep the layout even though this task has no I/O
        // typically threads with no I/O measurements.
        sprintf(dstr, " %9.2f %9.2f %9.2f", -1.0, -1.0, -1.0);
        cs->cprintf_s(cprint_support::IS_ZERO, (char *) "%s", dstr);
      }
      // I/O delays come from another file (/proc/#/stat)
      cs->cprintf_u64(cprint_support::NO_UNIT, 1, 7,
          (uint64_t) (pstc->blkio_swapin_delays - pstp->blkio_swapin_delays));
    }

    if(DISPLAY_CTXSW(actflag)){
      cs->cprintf_f(cprint_support::NO_UNIT, 2, 9, 2,
          s_value(pstp->nvcsw, pstc->nvcsw, itv), s_value(pstp->nivcsw, pstc->nivcsw, itv));
    }

    if(DISPLAY_KTAB(actflag)){
      cs->cprintf_u64(cprint_support::NO_UNIT, 1, 7, (uint64_t) pstc->threads);
      if(NO_PID_FD(plist->flags)){
        // /proc/#/fd directory not readable
        cs->cprintf_s(cprint_support::IS_ZERO, (char *) " %7s", (char *) "-1");
      }
      else{
        cs->cprintf_u64(cprint_support::NO_UNIT, 1, 7, (uint64_t) pstc->fd_nr);
      }
    }

    if(DISPLAY_RT(actflag)){
      cs->cprintf_u64(cprint_support::NO_UNIT, 1, 4, (uint64_t) pstc->priority);
      cs->cprintf_s(cprint_support::IS_STR, (char *) " %6s", (char *) GET_POLICY((unsigned int)pstc->policy));
    }

    print_comm(plist);
    again = 1;
  }
  delete cs;

  return again;
}


int write_pid_child_all_stats(int prev, int curr, int dis, char *prev_string, char *curr_string)
{
  multiplatform::cprint_support *cs = new multiplatform::cprint_support();
  struct pid_stats *pstc, *pstp;
  struct st_pid *plist;
  int again = 0;

  if(dis){
    PRINT_ID_HDR(prev_string, pidflag);
    if(DISPLAY_CPU(actflag)){ printf("    usr-ms system-ms  guest-ms"); }
    if(DISPLAY_MEM(actflag)){ printf(" minflt-nr majflt-nr"); }
    printf("  Command\n");
  }

  for(plist = pid_list; plist != NULL; plist = plist->next){
    if(get_pid_to_display(prev, curr, actflag, P_CHILD, plist) <= 0){ continue; }

    print_line_id(curr_string, plist);

    pstc = plist->pstats[curr];
    pstp = plist->pstats[prev];

    if(DISPLAY_CPU(actflag)){
      cs->cprintf_f(cprint_support::NO_UNIT, 3, 9, 0,
          (pstc->utime + pstc->cutime - pstc->gtime - pstc->cgtime) <
          (pstp->utime + pstp->cutime - pstp->gtime - pstp->cgtime) ?
          0.0 :
          (double) ((pstc->utime + pstc->cutime - pstc->gtime - pstc->cgtime) -
              (pstp->utime + pstp->cutime - pstp->gtime - pstp->cgtime)) / hz * 1000,
          (double) ((pstc->stime + pstc->cstime) - (pstp->stime + pstp->cstime)) / hz * 1000,
          (double) ((pstc->gtime + pstc->cgtime) - (pstp->gtime + pstp->cgtime)) / hz * 1000);
    }

    if(DISPLAY_MEM(actflag)){
      cs->cprintf_u64(cprint_support::NO_UNIT, 2, 9,
          (uint64_t) ((pstc->minflt + pstc->cminflt) - (pstp->minflt + pstp->cminflt)),
          (uint64_t) ((pstc->majflt + pstc->cmajflt) - (pstp->majflt + pstp->cmajflt)));
    }

    print_comm(plist);
    again = 1;
  }
  delete cs;

  return again;
}


int write_pid_task_cpu_stats(int prev, int curr, int dis, int disp_avg,
           char *prev_string, char *curr_string, uint64_t itv, uint64_t deltot_jiffies)
{
  multiplatform::cprint_support *cs = new multiplatform::cprint_support();
  struct pid_stats *pstc, *pstp;
  struct st_pid *plist;
  int again = 0;

  if(dis){
    PRINT_ID_HDR(prev_string, pidflag);
    printf("    %%usr %%system  %%guest   %%wait    %%CPU   CPU  Command\n");
  }

  for(plist = pid_list; plist != NULL; plist = plist->next){
    if(get_pid_to_display(prev, curr, P_A_CPU, P_TASK, plist) <= 0){ continue; }

    print_line_id(curr_string, plist);

    pstc = plist->pstats[curr];
    pstp = plist->pstats[prev];

    cs->cprintf_pc(DISPLAY_UNIT(pidflag), 5, 7, 2,
         (pstc->utime - pstc->gtime) < (pstp->utime - pstp->gtime) ?
         0.0 :
         sp_value(pstp->utime - pstp->gtime, pstc->utime - pstc->gtime, itv * hz / 100),
         sp_value(pstp->stime, pstc->stime, itv * hz / 100),
         sp_value(pstp->gtime, pstc->gtime, itv * hz / 100),
         sp_value(pstp->wtime, pstc->wtime, itv * hz / 100),
         // User time already includes guest time
         IRIX_MODE_OFF(pidflag) ?
         sp_value(pstp->utime + pstp->stime, pstc->utime + pstc->stime, deltot_jiffies) :
         sp_value(pstp->utime + pstp->stime, pstc->utime + pstc->stime, itv * hz / 100));

    if(!disp_avg){
      cs->cprintf_in(cprint_support::IS_INT, (char *) "   %3d", (char *) "", pstc->processor);
    }
    else{
      cs->cprintf_in(cprint_support::IS_STR, (char *) "%s", (char *) "     -", 0);
    }
    print_comm(plist);
    again = 1;
  }
  delete cs;

  return again;
}


int write_pid_child_cpu_stats(int prev, int curr, int dis, int disp_avg,
            char *prev_string, char *curr_string)
{
  multiplatform::cprint_support *cs = new multiplatform::cprint_support();
  struct pid_stats *pstc, *pstp;
  struct st_pid *plist;
  int rc, again = 0;

  if(dis){
    PRINT_ID_HDR(prev_string, pidflag);
    printf("    usr-ms system-ms  guest-ms  Command\n");
  }

  for(plist = pid_list; plist != NULL; plist = plist->next){
    // PID no longer exists
    if((rc = get_pid_to_display(prev, curr, P_A_CPU, P_CHILD, plist)) == 0){ continue; }

    // This will be used to compute average
    if(!disp_avg){ plist->uc_asum_count += 1; }

    if(rc < 0){ continue; } // PID should not be displayed

    print_line_id(curr_string, plist);

    pstc = plist->pstats[curr];
    pstp = plist->pstats[prev];

    if(disp_avg){
      cs->cprintf_f(cprint_support::NO_UNIT, 3, 9, 0,
          (pstc->utime + pstc->cutime - pstc->gtime - pstc->cgtime) <
          (pstp->utime + pstp->cutime - pstp->gtime - pstp->cgtime) ?
          0.0 :
          (double) ((pstc->utime + pstc->cutime - pstc->gtime - pstc->cgtime) -
              (pstp->utime + pstp->cutime - pstp->gtime - pstp->cgtime)) /
              (hz * plist->uc_asum_count) * 1000,
          (double) ((pstc->stime + pstc->cstime) -
              (pstp->stime + pstp->cstime)) /
              (hz * plist->uc_asum_count) * 1000,
          (double) ((pstc->gtime + pstc->cgtime) -
              (pstp->gtime + pstp->cgtime)) /
              (hz * plist->uc_asum_count) * 1000);
    }
    else{
      cs->cprintf_f(cprint_support::NO_UNIT, 3, 9, 0,
          (pstc->utime + pstc->cutime - pstc->gtime - pstc->cgtime) <
          (pstp->utime + pstp->cutime - pstp->gtime - pstp->cgtime) ?
          0.0 :
          (double) ((pstc->utime + pstc->cutime - pstc->gtime - pstc->cgtime) -
              (pstp->utime + pstp->cutime - pstp->gtime - pstp->cgtime)) / hz * 1000,
          (double) ((pstc->stime + pstc->cstime) - (pstp->stime + pstp->cstime)) / hz * 1000,
          (double) ((pstc->gtime + pstc->cgtime) - (pstp->gtime + pstp->cgtime)) / hz * 1000);
    }
    print_comm(plist);
    again = 1;
  }
  delete cs;

  return again;
}


int write_pid_task_memory_stats(int prev, int curr, int dis, int disp_avg,
        char *prev_string, char *curr_string, uint64_t itv)
{
  multiplatform::cprint_support *cs = new multiplatform::cprint_support();
  struct pid_stats *pstc, *pstp;
  struct st_pid *plist;
  int rc, again = 0;

  if(dis){
    PRINT_ID_HDR(prev_string, pidflag);
    printf("  minflt/s  majflt/s     VSZ     RSS   %%MEM  Command\n");
  }

  for(plist = pid_list; plist != NULL; plist = plist->next){
    // PID no longer exists
    if((rc = get_pid_to_display(prev, curr, P_A_MEM, P_TASK, plist)) == 0){ continue; }

    pstc = plist->pstats[curr];
    pstp = plist->pstats[prev];

    // This will be used to compute average
    if(!disp_avg){
      plist->total_vsz += pstc->vsz;
      plist->total_rss += pstc->rss;
      plist->rt_asum_count += 1;
    }

    if(rc < 0){ continue; } // PID should not be displayed

    print_line_id(curr_string, plist);

    cs->cprintf_f(cprint_support::NO_UNIT, 2, 9, 2,
        s_value(pstp->minflt, pstc->minflt, itv), s_value(pstp->majflt, pstc->majflt, itv));

    if(disp_avg){
      cs->cprintf_f(DISPLAY_UNIT(pidflag) ? cprint_support::UNIT_KILOBYTE : cprint_support::NO_UNIT, 2, 7, 0,
          (double) plist->total_vsz / plist->rt_asum_count,
          (double) plist->total_rss / plist->rt_asum_count);

      cs->cprintf_pc(DISPLAY_UNIT(pidflag), 1, 6, 2,
          tlmkb ? sp_value(0, plist->total_rss / plist->rt_asum_count, tlmkb) : 0.0);
    }
    else{
      // cast op. removed: uint64_t => uint64_t
      cs->cprintf_u64(DISPLAY_UNIT(pidflag) ? cprint_support::UNIT_KILOBYTE : cprint_support::NO_UNIT, 2, 7, pstc->vsz, pstc->rss);
      cs->cprintf_pc(DISPLAY_UNIT(pidflag), 1, 6, 2, tlmkb ? sp_value(0, pstc->rss, tlmkb) : 0.0);
    }

    print_comm(plist);
    again = 1;
  }
  delete cs;

  return again;
}


int write_pid_child_memory_stats(int prev, int curr, int dis, int disp_avg,
         char *prev_string, char *curr_string)
{
  multiplatform::cprint_support *cs = new multiplatform::cprint_support();
  struct pid_stats *pstc, *pstp;
  struct st_pid *plist;
  int rc, again = 0;

  if(dis){
    PRINT_ID_HDR(prev_string, pidflag);
    printf(" minflt-nr majflt-nr  Command\n");
  }

  for(plist = pid_list; plist != NULL; plist = plist->next){
    // PID no longer exists
    if((rc = get_pid_to_display(prev, curr, P_A_MEM, P_CHILD, plist)) == 0){ continue; }

    // This will be used to compute average
    if(!disp_avg){ plist->rc_asum_count += 1; }

    if(rc < 0){ continue; } // PID should not be displayed

    print_line_id(curr_string, plist);

    pstc = plist->pstats[curr];
    pstp = plist->pstats[prev];

    if(disp_avg){
      cs->cprintf_f(cprint_support::NO_UNIT, 2, 9, 0,
          (double) ((pstc->minflt + pstc->cminflt) -
              (pstp->minflt + pstp->cminflt)) / plist->rc_asum_count,
          (double) ((pstc->majflt + pstc->cmajflt) -
              (pstp->majflt + pstp->cmajflt)) / plist->rc_asum_count);
    }
    else{
      cs->cprintf_u64(cprint_support::NO_UNIT, 2, 9,
            (uint64_t) ((pstc->minflt + pstc->cminflt) - (pstp->minflt + pstp->cminflt)),
            (uint64_t) ((pstc->majflt + pstc->cmajflt) - (pstp->majflt + pstp->cmajflt)));
    }
    print_comm(plist);
    again = 1;
  }
  delete cs;

  return again;
}


int write_pid_stack_stats(int prev, int curr, int dis, int disp_avg,
        char *prev_string, char *curr_string)
{
  multiplatform::cprint_support *cs = new multiplatform::cprint_support();
  struct pid_stats *pstc;
  struct st_pid *plist;
  int rc, again = 0;

  if(dis){
    PRINT_ID_HDR(prev_string, pidflag);
    printf(" StkSize  StkRef  Command\n");
  }

  for(plist = pid_list; plist != NULL; plist = plist->next){
    // PID no longer exists
    if((rc = get_pid_to_display(prev, curr, P_A_STACK, P_NULL, plist)) == 0){ continue; }

    pstc = plist->pstats[curr];

    // This will be used to compute average
    if(!disp_avg){
      plist->total_stack_size += pstc->stack_size;
      plist->total_stack_ref  += pstc->stack_ref;
      plist->sk_asum_count += 1;
    }

    if(rc < 0){ continue; } // PID should not be displayed

    print_line_id(curr_string, plist);

    if(disp_avg){
      cs->cprintf_f(DISPLAY_UNIT(pidflag) ? cprint_support::UNIT_KILOBYTE : cprint_support::NO_UNIT, 2, 7, 0,
          (double) plist->total_stack_size / plist->sk_asum_count,
          (double) plist->total_stack_ref  / plist->sk_asum_count);
    }
    else{
      cs->cprintf_u64(DISPLAY_UNIT(pidflag) ? cprint_support::UNIT_KILOBYTE : cprint_support::NO_UNIT, 2, 7,
            (uint64_t) pstc->stack_size, (uint64_t) pstc->stack_ref);
    }

    print_comm(plist);
    again = 1;
  }
  delete cs;

  return again;
}


int write_pid_io_stats(int prev, int curr, int dis, int disp_avg,
           char *prev_string, char *curr_string, uint64_t itv)
{
  multiplatform::cprint_support *cs = new multiplatform::cprint_support();
  struct pid_stats *pstc, *pstp;
  struct st_pid *plist;
  char dstr[32];
  int rc, again = 0;
  double rbytes, wbytes, cbytes;

  if(dis){
    PRINT_ID_HDR(prev_string, pidflag);
    printf("   kB_rd/s   kB_wr/s kB_ccwr/s iodelay  Command\n");
  }

  for(plist = pid_list; plist != NULL; plist = plist->next){
    // PID no longer exists
    if((rc = get_pid_to_display(prev, curr, P_A_IO, P_NULL, plist)) == 0){ continue; }

    // This will be used to compute average delays
    if(!disp_avg){ plist->delay_asum_count += 1; }

    if(rc < 0){ continue; }// PID should not be displayed

    print_line_id(curr_string, plist);

    pstc = plist->pstats[curr];
    pstp = plist->pstats[prev];

    if(!NO_PID_IO(plist->flags)){
      rbytes = s_value(pstp->read_bytes,  pstc->read_bytes, itv);
      wbytes = s_value(pstp->write_bytes, pstc->write_bytes, itv);
      cbytes = s_value(pstp->cancelled_write_bytes, pstc->cancelled_write_bytes, itv);
      if(!DISPLAY_UNIT(pidflag)){ rbytes /= 1024; wbytes /= 1024; cbytes /= 1024; }
      cs->cprintf_f(DISPLAY_UNIT(pidflag) ? cprint_support::UNIT_BYTE : cprint_support::NO_UNIT, 3, 9, 2, rbytes, wbytes, cbytes);
    }
    else{
      // I/O file not readable (permission denied or file non existent)
      sprintf(dstr, " %9.2f %9.2f %9.2f", -1.0, -1.0, -1.0);
      cs->cprintf_s(cprint_support::IS_ZERO, (char *) "%s", dstr);
    }
    // I/O delays come from another file (/proc/#/stat)
    if(disp_avg){
      cs->cprintf_f(cprint_support::NO_UNIT, 1, 7, 0,
          (double) (pstc->blkio_swapin_delays - pstp->blkio_swapin_delays) /
          plist->delay_asum_count);
    }
    else{
      cs->cprintf_u64(cprint_support::NO_UNIT, 1, 7,
          (uint64_t) (pstc->blkio_swapin_delays - pstp->blkio_swapin_delays));
    }

    print_comm(plist);
    again = 1;
  }
  delete cs;

  return again;
}


int write_pid_ctxswitch_stats(int prev, int curr, int dis,
            char *prev_string, char *curr_string, uint64_t itv)
{
  multiplatform::cprint_support *cs = new multiplatform::cprint_support();
  struct pid_stats *pstc, *pstp;
  struct st_pid *plist;
  int again = 0;

  if(dis){
    PRINT_ID_HDR(prev_string, pidflag);
    printf("   cswch/s nvcswch/s  Command\n");
  }

  for(plist = pid_list; plist != NULL; plist = plist->next){
    if(get_pid_to_display(prev, curr, P_A_CTXSW, P_NULL, plist) <= 0){ continue; }

    print_line_id(curr_string, plist);

    pstc = plist->pstats[curr];
    pstp = plist->pstats[prev];

    cs->cprintf_f(cprint_support::NO_UNIT, 2, 9, 2,
        s_value(pstp->nvcsw,  pstc->nvcsw,  itv), s_value(pstp->nivcsw, pstc->nivcsw, itv));

    print_comm(plist);
    again = 1;
  }
  delete cs;

  return again;
}


int write_pid_rt_stats(int prev, int curr, int dis, char *prev_string, char *curr_string)
{
  multiplatform::cprint_support *cs = new multiplatform::cprint_support();
  struct pid_stats *pstc;
  struct st_pid *plist;
  int again = 0;

  if(dis){
    PRINT_ID_HDR(prev_string, pidflag);
    printf(" prio policy  Command\n");
  }

  for(plist = pid_list; plist != NULL; plist = plist->next){
    if(get_pid_to_display(prev, curr, P_A_RT, P_NULL, plist) <= 0){ continue; }

    print_line_id(curr_string, plist);

    pstc = plist->pstats[curr];

    cs->cprintf_u64(cprint_support::NO_UNIT, 1, 4, (uint64_t) pstc->priority);
    cs->cprintf_s(cprint_support::IS_STR, (char *) " %6s", (char *) GET_POLICY((unsigned int)pstc->policy));

    print_comm(plist);
    again = 1;
  }
  delete cs;

  return again;
}


int write_pid_ktab_stats(int prev, int curr, int dis, int disp_avg,
                         char *prev_string, char *curr_string)
{
  multiplatform::cprint_support *cs = new multiplatform::cprint_support();
  struct pid_stats *pstc;
  struct st_pid *plist;
  int rc, again = 0;

  if(dis){
    PRINT_ID_HDR(prev_string, pidflag);
    printf(" threads   fd-nr");
    printf("  Command\n");
  }

  for(plist = pid_list; plist != NULL; plist = plist->next){
    // PID no longer exists
    if((rc = get_pid_to_display(prev, curr, P_A_KTAB, P_NULL, plist)) == 0){ continue; }

    pstc = plist->pstats[curr];

    // This will be used to compute average
    if(!disp_avg){
      plist->total_threads += pstc->threads;
      plist->total_fd_nr   += pstc->fd_nr;
      plist->tf_asum_count += 1;
    }

    if(rc < 0){ continue; } // PID should not be displayed

    print_line_id(curr_string, plist);

    if(disp_avg){
      cs->cprintf_f(cprint_support::NO_UNIT, 2, 7, 0,
          (double) plist->total_threads / plist->tf_asum_count,
          NO_PID_FD(plist->flags) ? -1.0 : (double) plist->total_fd_nr / plist->tf_asum_count);
    }
    else{
      cs->cprintf_u64(cprint_support::NO_UNIT, 1, 7, (uint64_t) pstc->threads); // unsigned int => uint64_t
      if(NO_PID_FD(plist->flags)){
        cs->cprintf_s(cprint_support::IS_ZERO, (char *) " %7s", (char *) "-1");
      }
      else{
        cs->cprintf_u64(cprint_support::NO_UNIT, 1, 7, (uint64_t) pstc->fd_nr); // unsigned int => uint64_t
      }
    }

    print_comm(plist);
    again = 1;
  }
  delete cs;

  return again;
}


int write_stats_core(int prev, int curr, int dis, int disp_avg,
                     char *prev_string, char *curr_string)
{
  uint64_t itv, deltot_jiffies;
  int again = 0;

  // Test stdout
  {
    if(write(STDOUT_FILENO, "", 0) == -1){
      perror("stdout");
      exit(6);
    }
  }

  // Total number of jiffies spent on the interval
  deltot_jiffies = get_interval(tot_jiffies[prev], tot_jiffies[curr]);

  itv = get_interval(uptime_cs[prev], uptime_cs[curr]);

  if(DISPLAY_ONELINE(pidflag)){
    if(DISPLAY_TASK_STATS(tskflag)){
      again += write_pid_task_all_stats(prev, curr, dis, prev_string, curr_string, itv, deltot_jiffies);
    }
    if(DISPLAY_CHILD_STATS(tskflag)){
      again += write_pid_child_all_stats(prev, curr, dis, prev_string, curr_string);
    }
  }
  else{
    // Display CPU stats
    if(DISPLAY_CPU(actflag)){
      if(DISPLAY_TASK_STATS(tskflag)){
        again += write_pid_task_cpu_stats(prev, curr, dis, disp_avg, prev_string, curr_string, itv, deltot_jiffies);
      }
      if(DISPLAY_CHILD_STATS(tskflag)){
        again += write_pid_child_cpu_stats(prev, curr, dis, disp_avg, prev_string, curr_string);
      }
    }

    // Display memory stats
    if(DISPLAY_MEM(actflag)){
      if(DISPLAY_TASK_STATS(tskflag)){
        again += write_pid_task_memory_stats(prev, curr, dis, disp_avg, prev_string, curr_string, itv);
      }
      if(DISPLAY_CHILD_STATS(tskflag) && DISPLAY_MEM(actflag)){
        again += write_pid_child_memory_stats(prev, curr, dis, disp_avg, prev_string, curr_string);
      }
    }

    // Display stack stats
    if(DISPLAY_STACK(actflag)){
      again += write_pid_stack_stats(prev, curr, dis, disp_avg, prev_string, curr_string);
    }

    // Display I/O stats
    if(DISPLAY_IO(actflag)){
      again += write_pid_io_stats(prev, curr, dis, disp_avg, prev_string, curr_string, itv);
    }

    // Display context switches stats
    if(DISPLAY_CTXSW(actflag)){
      again += write_pid_ctxswitch_stats(prev, curr, dis, prev_string, curr_string, itv);
    }

    // Display kernel table stats
    if(DISPLAY_KTAB(actflag)){
      again += write_pid_ktab_stats(prev, curr, dis, disp_avg, prev_string, curr_string);
    }

    // Display scheduling priority and policy information
    if(DISPLAY_RT(actflag)){
      again += write_pid_rt_stats(prev, curr, dis, prev_string, curr_string);
    }
  }

  if(DISPLAY_ALL_PID(pidflag)){
    again = 1;
  }

  return again;
}


void write_stats_avg(int curr, int dis)
{
  constexpr static char string[] = "Average:";
  write_stats_core(2, curr, dis, TRUE, (char *) string, (char *) string);
}


int write_stats(int curr, int dis)
{
  constexpr static size_t TIMESTAMP_LEN = 64;
  char cur_time[2][TIMESTAMP_LEN];

  // Get previous timestamp
  if(DISPLAY_ONELINE(pidflag)){
    strcpy(cur_time[!curr], "# Time     ");
  }
  else if(PRINT_SEC_EPOCH(pidflag)){
    snprintf(cur_time[!curr], sizeof(cur_time[!curr]), "%-11ld", mktime(&ps_tstamp[!curr]));
    cur_time[!curr][sizeof(cur_time[!curr]) - 1] = '\0';
  }
  else if(is_iso_time_fmt()){
    strftime(cur_time[!curr], sizeof(cur_time[!curr]), "%H:%M:%S", &ps_tstamp[!curr]);
  }
  else{
    strftime(cur_time[!curr], sizeof(cur_time[!curr]), "%X", &ps_tstamp[!curr]);
  }

  // Get current timestamp
  if(PRINT_SEC_EPOCH(pidflag)){
    snprintf(cur_time[curr], sizeof(cur_time[curr]), "%-11ld", mktime(&ps_tstamp[curr]));
    cur_time[curr][sizeof(cur_time[curr]) - 1] = '\0';
  }
  else if(is_iso_time_fmt()){
    strftime(cur_time[curr], sizeof(cur_time[curr]), "%H:%M:%S", &ps_tstamp[curr]);
  }
  else{
    strftime(cur_time[curr], sizeof(cur_time[curr]), "%X", &ps_tstamp[curr]);
  }

  return (write_stats_core(!curr, curr, dis, FALSE, cur_time[!curr], cur_time[curr]));
}


void rw_pidstat_loop(int dis_hdr, int rows)
{
  int curr = 1, dis = 1;
  int again;
  uint64_t lines = rows;
  struct st_pid *plist;

  // Read system uptime
  {
    multiplatform::read_statistics *res = new multiplatform::read_statistics();
    res->read_uptime(&uptime_cs[0]);
    read_stats(0);
    delete res;
  }

  // Get total memory
  if(DISPLAY_MEM(actflag)){
    read_proc_meminfo();
  }

  // Display since boot time
  if(!interval && !EXEC_PGM(pidflag)){
    ps_tstamp[1] = ps_tstamp[0];
    write_stats(0, DISP_HDR);
    exit(0);
  }

  // Set a handler for SIGALRM
  memset(&alrm_act, 0, sizeof(alrm_act));
  alrm_act.sa_handler = alarm_handler;
  sigaction(SIGALRM, &alrm_act, NULL);
  if(interval){
    alarm(interval);
  }

  // Save the first stats collected. Will be used to compute the average
  ps_tstamp[2] = ps_tstamp[0];
  tot_jiffies[2] = tot_jiffies[0];
  uptime_cs[2] = uptime_cs[0];
  for(plist = pid_list; plist != NULL; plist = plist->next){
    memcpy(plist->pstats[2], plist->pstats[0], sizeof(struct pid_stats));
  }

  // Set a handler for SIGINT
  memset(&int_act, 0, sizeof(int_act));
  int_act.sa_handler = int_handler;
  sigaction(SIGINT, &int_act, NULL);

  // Wait for SIGALRM (or possibly SIGINT) signal
  pause();

  // SIGINT/SIGCHLD signals caught during first interval: Exit immediately
  if(signal_caught && interval){ return; }

  do {
    // Every PID is potentially nonexistent
    set_pid_nonexistent(pid_list);

    // Get time
    get_localtime(&ps_tstamp[curr], 0);

    // Read system uptime (in 1/100th of a second)
    multiplatform::read_statistics *res = new multiplatform::read_statistics();
    res->read_uptime(&(uptime_cs[curr]));
    delete res;

    // Read stats
    read_stats(curr);

    if(!dis_hdr){
      dis = lines / rows;
      if(dis){
        lines %= rows;
      }
      lines++;
    }

    // Print results
    again = write_stats(curr, dis);

    if(!again){ return; }

    if(count > 0){ count--; }

    if(count){
      pause();
      if(signal_caught){
        // SIGINT/SIGCHLD signals caught => Display average stats
        count = 0;
        printf("\n");  // Skip "^C" displayed on screen
      }
      else{
        curr ^= 1;
      }
    }
  }
  while (count);

  // The one line format uses a raw time value rather than time strings
  // so the average doesn't really fit.

  // Write stats average
  if(!DISPLAY_ONELINE(pidflag)){ write_stats_avg(curr, dis_hdr); }
}


pid_t exec_pgm(int argc, char **argv)
{
  pid_t child;
  char *args[argc + 1];
  int i;

  child = fork();

  switch(child){
    case -1:
      perror("fork");
      exit(4);
      break;

    case 0:
      // Child
      for(i = 0; i < argc; i++){ args[i] = argv[i]; }
      args[argc] = NULL;

      execvp(args[0], args);
      perror("exec");
      exit(4);
      break;

    default:
      // Parent.
      // Set a handler for SIGCHLD (signal that will be received
      // by pidstat when the child program terminates).
      // The handler is the same as for SIGINT: Stop and display average statistics.
      memset(&chld_act, 0, sizeof(chld_act));
      chld_act.sa_handler = int_handler;
      sigaction(SIGCHLD, &chld_act, NULL);

      return child;
  }
}

//////////
// from common.cpp
void print_version(void)
{
  constexpr static char VERSION[] = "0.1";
  printf("pidstat like application version %s\n", VERSION);
  printf("(C) Honda Research Institute Japan.\n");
  exit(0);
}


time_t get_localtime(struct tm *rectime, int d_off)
{
  constexpr static int SEC_PER_DAY = 3600 * 24;
  time_t timer;

  timer = time(NULL);
  timer -= SEC_PER_DAY * d_off;
  localtime_r(&timer, rectime);

  return timer;
}

/*
time_t get_gmtime(struct tm *rectime, int d_off)
{
  constexpr static int SEC_PER_DAY = 3600 * 24;
  time_t timer;

  timer = time(NULL);
  timer -= SEC_PER_DAY * d_off;
  gmtime_r(&timer, rectime);

  return timer;
}
*/

/*
time_t get_time(struct tm *rectime, int d_off)
{
  static int utc = 0;
  char *e;

  if(!utc){
    // Read environment variable value once
    if((e = getenv("S_TIME_DEF_TIME")) != NULL){ utc = !strcmp(e, K_UTC); }
    utc++;
  }

  if(utc == 2){ return get_gmtime(rectime, d_off); }
  else{         return get_localtime((struct tm *) rectime, (int) d_off); }
}
*/









// ***************************************************************************
// * Main entry to the pidstat program.
// ***************************************************************************
int main(int argc, char **argv)
{
  int opt = 1, dis_hdr = -1, pid_nr = 0;
  int i;
  unsigned int pid;
  struct utsname header;
  int rows = 23;
  char *t;

  // Init color strings
  //{
  //  multiplatform::cprint_support *cs = new multiplatform::cprint_support();
  //  init_colors();
  //  delete cs;
  //}

  // Get HZ
  {
    multiplatform::read_statistics *res = new multiplatform::read_statistics();
    res->get_HZ();
    hz = res->_hz;
    delete res;
  }

  // Compute page shift in kB
  //{
  //  multiplatform::read_statistics *res = new multiplatform::read_statistics();
  //  res->get_kb_shift();
  //  delete res;
  //}

  // Process args...
  while (opt < argc){
    if(!strcmp(argv[opt], "-e")){
      if(!argv[++opt]){ usage(argv[0]); }
      pidflag |= P_D_PID + P_F_EXEC_PGM;
      add_list_pid(&pid_list, exec_pgm(argc - opt, argv + opt), 0);
      break;
    }
    else if(!strcmp(argv[opt], "-p")){
      pidflag |= P_D_PID;
      if(!argv[++opt]){ usage(argv[0]); }
      for(t = strtok(argv[opt], ","); t; t = strtok(NULL, ","), pid_nr++){
        if(!strcmp(t, "ALL")){ pidflag |= P_D_ALL_PID; pid_nr++; }
        else if(!strcmp(t, "SELF")){ add_list_pid(&pid_list, getpid(), 0); }
        else{
          if(strspn(t, "0123456789") != strlen(t)){ usage(argv[0]); }
          pid = atoi(t);
          if(pid < 1){ usage(argv[0]); }
          add_list_pid(&pid_list, pid, 0);
        }
      }
      opt++;
    }
    else if(!strcmp(argv[opt], "-C")){
      if(!argv[++opt]){ usage(argv[0]); }
      strncpy(commstr, argv[opt++], sizeof(commstr));
      commstr[sizeof(commstr) - 1] = '\0';
      pidflag |= P_F_COMMSTR;
      if(!strlen(commstr)){ usage(argv[0]); }
    }
    else if(!strcmp(argv[opt], "-G")){
      if(!argv[++opt]){ usage(argv[0]); }
      strncpy(procstr, argv[opt++], sizeof(procstr));
      procstr[sizeof(procstr) - 1] = '\0';
      pidflag |= P_F_PROCSTR;
      if(!strlen(procstr)){ usage(argv[0]); }
    }
    else if(!strcmp(argv[opt], "--human")){
      pidflag |= P_D_UNIT;
      opt++;
    }
    else if(!strncmp(argv[opt], "--dec=", 6) && (strlen(argv[opt]) == 7)){
      // Get number of decimal places
      dplaces_nr = atoi(argv[opt] + 6);
      if((dplaces_nr < 0) || (dplaces_nr > 2)){ usage(argv[0]); }
      opt++;
    }
    else if(!strcmp(argv[opt], "-T")){
      if(!argv[++opt]){ usage(argv[0]); }
      if(tskflag){ dis_hdr++; }
      if(!strcmp(argv[opt], "TASK")){ tskflag |= P_TASK; }
      else if(!strcmp(argv[opt], "CHILD")){ tskflag |= P_CHILD; }
      else if(!strcmp(argv[opt], "ALL")){ tskflag |= P_TASK + P_CHILD; dis_hdr++; }
      else{ usage(argv[0]); }
      opt++;
    }

    // Option used individually. See below for grouped option
    else if(!strcmp(argv[opt], "-U")){
      // Display username instead of UID
      pidflag |= P_D_USERNAME;
      if(argv[++opt] && (argv[opt][0] != '-') &&
                        (strspn(argv[opt], "0123456789") != strlen(argv[opt]))){
        strncpy(userstr, argv[opt++], sizeof(userstr));
        userstr[sizeof(userstr) - 1] = '\0';
        pidflag |= P_F_USERSTR;
        if(!strlen(userstr)){ usage(argv[0]); }
      }
    }

    else if(!strncmp(argv[opt], "-", 1)){
      for(i = 1; *(argv[opt] + i); i++){
        switch(*(argv[opt] + i)){
        case 'd': actflag |= P_A_IO; dis_hdr++; break; // Display I/O usage
        case 'H': pidflag |= P_D_SEC_EPOCH; break; // Display timestamps in sec since the epoch
        case 'h': pidflag |= P_D_ONELINE; break; // Display stats on one line
        case 'I': pidflag |= P_F_IRIX_MODE; break; // IRIX mode off
        case 'l': pidflag |= P_D_CMDLINE; break; // Display whole command line
        case 'R': actflag |= P_A_RT; dis_hdr++; break; // Display priority and policy info
        case 'r': actflag |= P_A_MEM; dis_hdr++; break; // Display memory usage
        case 's': actflag |= P_A_STACK; dis_hdr++; break; // Display stack sizes
        case 't': pidflag |= P_D_TID; break; // Display stats for threads
        case 'U': pidflag |= P_D_USERNAME; break; // When option is grouped, it cannot take an arg
        case 'u': actflag |= P_A_CPU; dis_hdr++; break; // Display CPU usage
        case 'V': print_version(); break; // Print version number and exit
        case 'v': actflag |= P_A_KTAB; dis_hdr++; break; // Display some kernel tables values
        case 'w': actflag |= P_A_CTXSW; dis_hdr++; break; // Display context switches
        default: usage(argv[0]);
        }
      }
      opt++;
    }

    else if(interval < 0){ // Get interval
      if(strspn(argv[opt], "0123456789") != strlen(argv[opt])){ usage(argv[0]); }
      interval = atol(argv[opt++]);
      if(interval < 0){ usage(argv[0]); }
      count = -1;
    }
    else if(count <= 0){ // Get count value
      if((strspn(argv[opt], "0123456789") != strlen(argv[opt])) || !interval){ usage(argv[0]); }
      count = atol(argv[opt++]);
      if(count < 1){ usage(argv[0]); }
    }
    else{
      usage(argv[0]);
    }
  }

  // Interval not set => display stats since boot time
  if(interval < 0){ interval = 0; }

  if(!DISPLAY_PID(pidflag)){ dis_hdr = 1; }

  // Check flags and set default values
  check_flags();

  // Count nb of proc
  //cpu_nr = get_cpu_nr(~0, FALSE);
  {
    multiplatform::read_statistics *res = new multiplatform::read_statistics();
    cpu_nr = res->get_cpu_nr(~0, FALSE);
    delete res;
  }

  if(dis_hdr < 0){ dis_hdr = 0; }
  if(!dis_hdr){
    if((pid_nr > 1)){ dis_hdr = 1; }
    else{ rows = get_win_height(); }
  }

  // Get time
  get_localtime(&(ps_tstamp[0]), 0);

  // Don't buffer data if redirected to a pipe.
  // Note: With musl-c, the behavior of this function is undefined except
  // when it is the first operation on the stream.
  setbuf(stdout, NULL);

  // Get system name, release number and hostname
  uname(&header);
  print_gal_header(&(ps_tstamp[0]), header.sysname, header.release,
                     header.nodename, header.machine, cpu_nr, PLAIN_OUTPUT);

  // Main loop
  rw_pidstat_loop(dis_hdr, rows);

  // Free structures
  sfree_pid(&pid_list, TRUE);

  return 0;
}

} // namespace multiplatform

int main(int argc, char **argv)
{
  return multiplatform::main(argc, argv);
}

