// read_stats.cpp: Read system statistics

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <errno.h>
#include <dirent.h>
#include <ctype.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/statvfs.h>
#include <unistd.h>
#include <inttypes.h>

#include "stat_print_base.hpp"
#include "read_stats.hpp"


namespace multiplatform {

int read_statistics::read_stat_cpu(struct stats_cpu *st_cpu, int nr_alloc)
{
  constexpr static char STAT[] = "/proc/stat";
  FILE *fp;
  struct stats_cpu *st_cpu_i;
  struct stats_cpu sc;
  char line[8192];
  int proc_nr;
  int cpu_read = 0;

  if((fp = fopen(STAT, "r")) == NULL){
    fprintf(stderr, "Cannot open %s: %s\n", STAT, strerror(errno));
    exit(2);
  }

  while (fgets(line, sizeof(line), fp) != NULL){
    if(!strncmp(line, "cpu ", 4)){
      // All the fields don't necessarily exist,
      // depending on the kernel version used.
      memset(st_cpu, 0, sizeof(struct stats_cpu));

      // Read the number of jiffies spent in the different modes
      // (user, nice, etc.) among all proc. CPU usage is not reduced
      // to one processor to avoid rounding problems.
      sscanf(line + 5, "%" PRIu64 " %" PRIu64 " %" PRIu64 " %" PRIu64 " %" PRIu64 " %" PRIu64 " %" PRIu64 " %" PRIu64 " %" PRIu64 " %" PRIu64,
             &st_cpu->cpu_user, &st_cpu->cpu_nice,
             &st_cpu->cpu_sys, &st_cpu->cpu_idle, &st_cpu->cpu_iowait,
             &st_cpu->cpu_hardirq, &st_cpu->cpu_softirq, &st_cpu->cpu_steal,
             &st_cpu->cpu_guest, &st_cpu->cpu_guest_nice);

      if(!cpu_read){ cpu_read = 1; }

      if(nr_alloc == 1){ break; } // We just want to read stats for CPU "all"
    }

    else if(!strncmp(line, "cpu", 3)){
      // All the fields don't necessarily exist
      memset(&sc, 0, sizeof(struct stats_cpu));
      // Read the number of jiffies spent in the different modes
      // (user, nice, etc) for current proc.
      // This is done only on SMP machines.
      sscanf(line + 3, "%d %" PRIu64 " %" PRIu64 " %" PRIu64 " %" PRIu64 " %" PRIu64 " %" PRIu64 " %" PRIu64 " %" PRIu64 " %" PRIu64 " %" PRIu64,
             &proc_nr,
             &sc.cpu_user, &sc.cpu_nice, &sc.cpu_sys, &sc.cpu_idle, &sc.cpu_iowait,
             &sc.cpu_hardirq, &sc.cpu_softirq, &sc.cpu_steal,
             &sc.cpu_guest, &sc.cpu_guest_nice);

      if(proc_nr + 2 > nr_alloc){
        cpu_read = -1;
        break;
      }

      st_cpu_i = st_cpu + proc_nr + 1;
      *st_cpu_i = sc;

      if(proc_nr + 2 > cpu_read){ cpu_read = proc_nr + 2; }
    }
  }

  fclose(fp);
  return cpu_read;
}

int read_statistics::read_meminfo(struct stats_memory *st_memory)
{
  constexpr static char MEMINFO[] = "/proc/meminfo";
  FILE *fp;
  char line[128];

  if((fp = fopen(MEMINFO, "r")) == NULL){ return 0; }

  while (fgets(line, sizeof(line), fp) != NULL){
    if(!strncmp(line, "MemTotal:", 9)){
      // Read the total amount of memory in kB
      sscanf(line + 9, "%" PRIu64, &st_memory->tlmkb);
    }
    else if(!strncmp(line, "MemFree:", 8)){
      // Read the amount of free memory in kB
      sscanf(line + 8, "%" PRIu64, &st_memory->frmkb);
    }
    else if(!strncmp(line, "MemAvailable:", 13)){
      // Read the amount of available memory in kB
      sscanf(line + 13, "%" PRIu64, &st_memory->availablekb);
    }
    else if(!strncmp(line, "Buffers:", 8)){
      // Read the amount of buffered memory in kB
      sscanf(line + 8, "%" PRIu64, &st_memory->bufkb);
    }
    else if(!strncmp(line, "Cached:", 7)){
      // Read the amount of cached memory in kB
      sscanf(line + 7, "%" PRIu64, &st_memory->camkb);
    }
    else if(!strncmp(line, "SwapCached:", 11)){
      // Read the amount of cached swap in kB
      sscanf(line + 11, "%" PRIu64, &st_memory->caskb);
    }
    else if(!strncmp(line, "Active:", 7)){
      // Read the amount of active memory in kB
      sscanf(line + 7, "%" PRIu64, &st_memory->activekb);
    }
    else if(!strncmp(line, "Inactive:", 9)){
      // Read the amount of inactive memory in kB
      sscanf(line + 9, "%" PRIu64, &st_memory->inactkb);
    }
    else if(!strncmp(line, "SwapTotal:", 10)){
      // Read the total amount of swap memory in kB
      sscanf(line + 10, "%" PRIu64, &st_memory->tlskb);
    }
    else if(!strncmp(line, "SwapFree:", 9)){
      // Read the amount of free swap memory in kB
      sscanf(line + 9, "%" PRIu64, &st_memory->frskb);
    }
    else if(!strncmp(line, "Dirty:", 6)){
      // Read the amount of dirty memory in kB
      sscanf(line + 6, "%" PRIu64, &st_memory->dirtykb);
    }
    else if(!strncmp(line, "Committed_AS:", 13)){
      // Read the amount of commited memory in kB
      sscanf(line + 13, "%" PRIu64, &st_memory->comkb);
    }
    else if(!strncmp(line, "AnonPages:", 10)){
      // Read the amount of pages mapped into userspace page tables in kB
      sscanf(line + 10, "%" PRIu64, &st_memory->anonpgkb);
    }
    else if(!strncmp(line, "Slab:", 5)){
      // Read the amount of in-kernel data structures cache in kB
      sscanf(line + 5, "%" PRIu64, &st_memory->slabkb);
    }
    else if(!strncmp(line, "KernelStack:", 12)){
      // Read the kernel stack utilization in kB
      sscanf(line + 12, "%" PRIu64, &st_memory->kstackkb);
    }
    else if(!strncmp(line, "PageTables:", 11)){
      // Read the amount of memory dedicated to the lowest level of page tables in kB
      sscanf(line + 11, "%" PRIu64, &st_memory->pgtblkb);
    }
    else if(!strncmp(line, "VmallocUsed:", 12)){
      // Read the amount of vmalloc area which is used in kB
      sscanf(line + 12, "%" PRIu64, &st_memory->vmusedkb);
    }
  }

  fclose(fp);
  return 1;
}


void read_statistics::read_uptime(uint64_t *uptime)
{
  constexpr static char UPTIME[] = "/proc/uptime";
  FILE *fp = NULL;
  char line[128];
  uint64_t up_sec, up_cent;
  int err = FALSE;

  if((fp = fopen(UPTIME, "r")) == NULL){ err = TRUE; }
  else if(fgets(line, sizeof(line), fp) == NULL){ err = TRUE; }
  else if(sscanf(line, "%lu.%lu", &up_sec, &up_cent) == 2){
    *uptime = (uint64_t) up_sec * 100 + (uint64_t) up_cent;
  }
  else{ err = TRUE; }

  if(fp != NULL){ fclose(fp); }
  if(err){
    fprintf(stderr, "Cannot read %s\n", UPTIME);
    exit(2);
  }
}



















unsigned int read_statistics::get_sys_cpu_nr(int highest)
{
  constexpr static char SYSFS_DEVCPU[] = "/sys/devices/system/cpu";
  DIR *dir;
  struct dirent *drd;
  struct stat buf;
  char line[512]; // SYSFS_DEVCPU[24] + "/"[1] + d_name[256] + "\0"[1]
  unsigned int num_cpu, cpu_nr = 0;

  if((dir = opendir(SYSFS_DEVCPU)) == NULL){ return 0; }

  line[sizeof(line) - 1] = '\0'; // To make the code more secure.
  if(highest){
    while((drd = readdir(dir)) != NULL){
      if(!strncmp(drd->d_name, "cpu", 3) && isdigit(drd->d_name[3])){
        snprintf(line, sizeof(line) - 1, "%s/%s", SYSFS_DEVCPU, drd->d_name);
        if(stat(line, &buf) < 0){ continue; }
        if(S_ISDIR(buf.st_mode)){
          sscanf(drd->d_name + 3, "%d", &num_cpu);
          if(num_cpu > cpu_nr){ cpu_nr = num_cpu; }
        }
      }
    }
    ++cpu_nr;
  }
  else{
    while((drd = readdir(dir)) != NULL){
      if(!strncmp(drd->d_name, "cpu", 3) && isdigit(drd->d_name[3])){
        snprintf(line, sizeof(line) - 1, "%s/%s", SYSFS_DEVCPU, drd->d_name);
        if(stat(line, &buf) < 0){ continue; }
        if(S_ISDIR(buf.st_mode)){ ++cpu_nr; }
      }
    }
  }

  closedir(dir);

  return cpu_nr;
}


unsigned int read_statistics::get_proc_cpu_nr(void)
{
  constexpr static char STAT[] = "/proc/stat";
  FILE *fp;
  char line[16];
  unsigned int num_cpu, cpu_nr = 0;

  if((fp = fopen(STAT, "r")) == NULL){
    fprintf(stderr, "Cannot open %s: %s\n", STAT, strerror(errno));
    return 0;
  }

  while(fgets(line, sizeof(line), fp) != NULL){
    if(strncmp(line, "cpu ", 4) && !strncmp(line, "cpu", 3)){
      sscanf(line + 3, "%d", &num_cpu);
      if(num_cpu > cpu_nr){ cpu_nr = num_cpu; }
    }
  }
  ++cpu_nr;

  fclose(fp);

  return cpu_nr;
}


unsigned int read_statistics::get_cpu_nr(unsigned int max_nr_cpus, int highest)
{
  unsigned int cpu_nr;

  if((cpu_nr = get_sys_cpu_nr(highest)) == 0){
    cpu_nr = get_proc_cpu_nr();
  }

  if(cpu_nr > max_nr_cpus){
    fprintf(stderr, "Cannot handle so many processors!\n");
    return 0;
  }

  return cpu_nr;
}












void read_statistics::get_HZ(void)
{
  long ticks;

  if((ticks = sysconf(_SC_CLK_TCK)) == -1){ perror("sysconf"); }

  _hz = (uint64_t) ticks;
}


void read_statistics::get_kb_shift(void)
{
  int shift = 0;
  long size;

  // One can also use getpagesize() to get the size of a page
  if((size = sysconf(_SC_PAGESIZE)) == -1){ perror("sysconf"); }

  size >>= 10;  // Assume that a page has a minimum size of 1 kB

  while(size > 1){ shift++; size >>= 1; }

  _kb_shift = (unsigned int) shift;
}









} // namespace multiplatform

