// statistics print base

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <stdarg.h>
#include <inttypes.h>
#include <time.h>
#include <errno.h>
#include <unistd.h>  // For STDOUT_FILENO, among others
#include <sys/ioctl.h>
#include <sys/types.h>
#include <dirent.h>
#include <ctype.h>
#include <limits.h>
#include <libgen.h>

#include "stat_print_base.hpp"

namespace multiplatform {

// Number of decimal places
extern int dplaces_nr;

int is_iso_time_fmt(void)
{
  static int is_iso = -1;
  char *e;

  if(is_iso < 0){ is_iso = (((e = getenv("S_TIME_FORMAT")) != NULL) && !strcmp(e, K_ISO)); }
  return is_iso;
}


void xprintf(int nr_tab, const char *fmtf, ...)
{
  static char buf[1024];
  va_list args;

  va_start(args, fmtf);
  vsnprintf(buf, sizeof(buf), fmtf, args);
  va_end(args);

  //prtab(nr_tab);
  { int i; for(i = 0; i < nr_tab; i++){ printf("\t"); } }
  printf("%s\n", buf);
}


int print_gal_header(struct tm *rectime, char *sysname, char *release,
		     char *nodename, char *machine, int cpu_nr, int format)
{
  int rc = 0;

  std::string tmp_date;
  {
    if(rectime == NULL){
      tmp_date = "?/?/?";
    }
    else if(is_iso_time_fmt()){
      char tmp[64];
      strftime(tmp, 64, "%Y-%m-%d", rectime);
      tmp_date = std::string(tmp);
      rc = 1;
    }
    else{
      char tmp[64];
      strftime(tmp, 64, "%x", rectime);
      tmp_date = std::string(tmp);
    }
  }

  if(format == PLAIN_OUTPUT){
    printf("%s %s (%s) \t%s \t_%s_\t(%d CPU)\n",
           sysname, release, nodename, tmp_date.c_str(), machine, cpu_nr);
  }
  else{ // JSON output
    xprintf(0, "{\"sysstat\": {");
    xprintf(1, "\"hosts\": [");
    xprintf(2, "{");
    xprintf(3, "\"nodename\": \"%s\",", nodename);
    xprintf(3, "\"sysname\": \"%s\",", sysname);
    xprintf(3, "\"release\": \"%s\",", release);
    xprintf(3, "\"machine\": \"%s\",", machine);
    xprintf(3, "\"number-of-cpus\": %d,", cpu_nr);
    xprintf(3, "\"date\": \"%s\",", tmp_date.c_str());
    xprintf(3, "\"statistics\": [");
  }

  return rc;
}


int get_win_height(void)
{
  struct winsize win;
  int rows = 3600 * 24; // This defalut value use for the case of file or pipe, etc...

  if(ioctl(STDOUT_FILENO, TIOCGWINSZ, &win) != -1){
    if(win.ws_row > 2){ rows = win.ws_row - 2; }
  }
  return rows;
}


// prev_time=0 when displaying stats since system startup
uint64_t get_interval(uint64_t prev_uptime, uint64_t curr_uptime)
{
  return (prev_uptime >= curr_uptime) ? 1 : (curr_uptime - prev_uptime); // Minimum is 1.
}




////////// cprint_support class start //////////

void cprint_support::init_colors(void)
{
  constexpr static char NEVER[] = "never";
  constexpr static char ALWAYS[] = "always";
  char *e, *p;
  size_t len;

  if((e = getenv("S_COLORS")) == NULL
       ? !isatty(STDOUT_FILENO)
       : (!strcmp(e, NEVER) || (strcmp(e, ALWAYS) && !isatty(STDOUT_FILENO)))) {
    // Environment variable not set and stdout is not a terminal,
    // or set to "never", or set to "auto" and stdout is not a terminal: Unset color strings.
    strcpy(sc_percent_high, "");
    strcpy(sc_percent_low, "");
    strcpy(sc_zero_int_stat, "");
    strcpy(sc_int_stat, "");
    strcpy(sc_item_name, "");
    strcpy(sc_sa_comment, "");
    strcpy(sc_sa_restart, "");
    strcpy(sc_normal, "");

    return;
  }

  if((e = getenv("S_COLORS_SGR")) == NULL){ return; } // Environment variable not set

  for(p = strtok(e, ":"); p; p = strtok(NULL, ":")){
    len = strlen(p);
    if((len > 7) || (len < 3) || (*(p + 1) != '=') ||
        (strspn(p + 2, ";0123456789") != (len - 2))){ continue; } // Ignore malformed codes

    switch(*p){
    case 'H': snprintf(sc_percent_high,  MAX_SGR_LEN, "\e[%sm", p + 2); break;
    case 'M': snprintf(sc_percent_low,   MAX_SGR_LEN, "\e[%sm", p + 2); break;
    case 'Z': snprintf(sc_zero_int_stat, MAX_SGR_LEN, "\e[%sm", p + 2); break;
    case 'N': snprintf(sc_int_stat,      MAX_SGR_LEN, "\e[%sm", p + 2); break;
    case 'I': snprintf(sc_item_name,     MAX_SGR_LEN, "\e[%sm", p + 2); break;
    case 'C': snprintf(sc_sa_comment,    MAX_SGR_LEN, "\e[%sm", p + 2); break;
    case 'R': snprintf(sc_sa_restart,    MAX_SGR_LEN, "\e[%sm", p + 2); break;
    }
  }
}


void cprint_support::cprintf_unit(int unit, int wi, double dval)
{
  // Units (sectors, Bytes, kilobytes, etc.)
  constexpr char units[] = {'s', 'B', 'k', 'M', 'G', 'T', 'P', '?'};

  if(wi < 4){ wi = 4; } // E.g. 1.3M
  if(!unit){
    // Value is a number of sectors. Convert it to kB
    dval /= 2;
    unit = 2;
  }
  while(dval >= 1024){
    dval /= 1024;
    unit++;
  }
  printf(" %*.*f", wi - 1, dplaces_nr ? 1 : 0, dval);
  printf("%s", sc_normal);

  // Display unit
  if(unit >= NR_UNITS){
    unit = NR_UNITS - 1;
  }
  printf("%c", units[unit]);
}


void cprint_support::cprintf_u64(int unit, int num, int wi, ...)
{
  int i;
  uint64_t val;
  va_list args;

  va_start(args, wi);

  for(i = 0; i < num; i++){
    val = va_arg(args, uint64_t);
    if(!val){ printf("%s", sc_zero_int_stat); }
    else{ printf("%s", sc_int_stat); }
    if(unit < 0){
      printf(" %*" PRIu64, wi, val);
      printf("%s", sc_normal);
    }
    else{
      cprintf_unit(unit, wi, (double) val);
    }
  }

  va_end(args);
}


void cprint_support::cprintf_x(int num, int wi, ...)
{
  int i;
  unsigned int val;
  va_list args;

  va_start(args, wi);

  for(i = 0; i < num; i++){
    val = va_arg(args, unsigned int);
    printf("%s", sc_int_stat);
    printf(" %*x", wi, val);
    printf("%s", sc_normal);
  }

  va_end(args);
}


void cprint_support::cprintf_f(int unit, int num, int wi, int wd, ...)
{
  int i;
  double val, lim = 0.005;
  va_list args;

  // If there are decimal places then get the value entered on the command line (if existing).
  if((wd > 0) && (dplaces_nr >= 0)){ wd = dplaces_nr; }

  // Update limit value according to number of decimal places
  if(wd == 1){ lim = 0.05; }

  va_start(args, wd);

  for(i = 0; i < num; i++){
    val = va_arg(args, double);
    if(((wd > 0) && (val < lim) && (val > (lim * -1))) ||
       ((wd == 0) && (val <= 0.5) && (val >= -0.5))){  // "Round half to even" law
      printf("%s", sc_zero_int_stat);
    }
    else{
      printf("%s", sc_int_stat);
    }

    if(unit < 0){
      printf(" %*.*f", wi, wd, val);
      printf("%s", sc_normal);
    }
    else{
      cprintf_unit(unit, wi, val);
    }
  }

  va_end(args);
}


void cprint_support::cprintf_pc(int human, int num, int wi, int wd, ...)
{
  constexpr static double PERCENT_LIMIT_HIGH = 75.0;
  constexpr static double PERCENT_LIMIT_LOW = 50.0;
  int i;
  double val, lim = 0.005;
  va_list args;

  // If there are decimal places then get the value entered on the command line (if existing).
  if((wd > 0) && (dplaces_nr >= 0)){ wd = dplaces_nr; }

  // If a percent sign is to be displayed, then there will be zero (or one) decimal place.
  if(human > 0){
    if(wi < 4){ wi = 4; } // E.g., 100%
    // Keep one place for the percent sign
    wi -= 1;
    if(wd > 1){ wd -= 1; }
  }

  // Update limit value according to number of decimal places
  if(wd == 1){ lim = 0.05; }

  va_start(args, wd);

  for(i = 0; i < num; i++){
    val = va_arg(args, double);
    if(val >= PERCENT_LIMIT_HIGH){ printf("%s", sc_percent_high); }
    else if(val >= PERCENT_LIMIT_LOW){ printf("%s", sc_percent_low); }
    else if(((wd > 0) && (val < lim)) || ((wd == 0) && (val <= 0.5))){
      // "Round half to even" law
      printf("%s", sc_zero_int_stat);
    }
    else{ printf("%s", sc_int_stat); }
    printf(" %*.*f", wi, wd, val);
    printf("%s", sc_normal);
    if(human > 0){ printf("%%"); }
  }

  va_end(args);
}


void cprint_support::cprintf_in(int type, char *format, char *item_string, int item_int)
{
  printf("%s", sc_item_name);
  if(type){ printf(format, item_string); }
  else{ printf(format, item_int); }
  printf("%s", sc_normal);
}


void cprint_support::cprintf_s(int type, char *format, char *string)
{
  if(type == IS_STR){ printf("%s", sc_int_stat); }
  else if(type == IS_ZERO){ printf("%s", sc_zero_int_stat); }
  else if(type == IS_RESTART){ printf("%s", sc_sa_restart); }
  else{ printf("%s", sc_sa_comment); } // IS_COMMENT
  printf(format, string);
  printf("%s", sc_normal);
}

////////// cprint_support class end //////////

} // namespace multiplatform

